"""Physical fields output (:mod:`fluidsim.solvers.ns2d.strat.output.phys_fields`)
================================================================================

Provides:

.. autoclass:: PhysFields2DStrat
   :members:
   :private-members:

"""

from fluidsim.base.output.phys_fields2d import PhysFieldsBase2D


class PhysFields2DStrat(PhysFieldsBase2D):
    """Class physical fields of solver ns2d.strat """

    def update_animation(self, frame, **fargs):

        super().update_animation(frame, **fargs)

        if hasattr(self, "ratio_omegas"):
            title = (
                self.key_field
                + f", R = {self.output.ratio_omegas:.2f}"
                + f", F = {self.output.froude_number:.1f}"
                + f", $t = {self.ani_times[frame]:.3f}$"
            )
        else:
            title = self.key_field + f", $t = {self.ani_times[frame]:.3f}$"

        self.ax.set_title(title)



#!/usr/bin/env python
# coding=utf8

import numpy as np
import matplotlib.pyplot as plt

from solveq2d import solveq2d

from create_figs_articles import CreateFigArticles


num_fig = 1000
SAVE_FIG = 1
name_file = "spectra_forcingq.eps"

create_fig = CreateFigArticles(
    short_name_article="SW1l", SAVE_FIG=SAVE_FIG, FOR_BEAMER=False, fontsize=19
)


name_dir_results = (
    create_fig.path_base_dir + "/Results_for_article_SW1l"
    "/Approach_runs_2048x2048"
    "/SE2D_SW1lexlin_forcing_L=50.x50._2048x2048_c2=400_f=0_2013-05-29_23-54-57"
)

sim = solveq2d.create_sim_plot_from_dir(name_dir_results)
tmin = 30
dict_results = sim.output.spectra.load2D_mean(tmin=tmin)

kh = dict_results["kh"]

EK = dict_results["spectrum2D_EK"]
EA = dict_results["spectrum2D_EA"]
EKr = dict_results["spectrum2D_EKr"]
E_tot = EK + EA
EKd = EK - EKr
Edlin = dict_results["spectrum2D_Edlin"]

fig, ax1 = create_fig.figure_axe(name_file=name_file)
ax1.set_xscale("log")
ax1.set_yscale("log")

coef_compensate = 5.0 / 3
coef_norm = kh ** coef_compensate

l_Etot = ax1.plot(kh, E_tot * coef_norm, "k", linewidth=4)
l_EK = ax1.plot(kh, EK * coef_norm, "r", linewidth=2)
l_EA = ax1.plot(kh, EA * coef_norm, "b", linewidth=2)
ax1.plot(kh, EKr * coef_norm, "r--", linewidth=2)
ax1.plot(kh, EKd * coef_norm, "r:", linewidth=2)

ax1.plot(kh, -EK * coef_norm, "m", linewidth=2)
ax1.plot(kh, -EKd * coef_norm, "m:", linewidth=2)

ax1.plot(kh, Edlin * coef_norm, "y:", linewidth=2)


cond = np.logical_and(kh > 1, kh < 20)
ax1.plot(kh[cond], 1e1 * kh[cond] ** (-3.0) * coef_norm[cond], "k--", linewidth=1)
plt.figtext(0.6, 0.78, "$k^{-3}$", fontsize=20)

cond = np.logical_and(kh > 0.3, kh < 10)
ax1.plot(kh[cond], 4e-2 * kh[cond] ** (-2.0) * coef_norm[cond], "k:", linewidth=1)
plt.figtext(0.25, 0.55, "$k^{-2}$", fontsize=20)

cond = np.logical_and(kh > 0.3, kh < 15)
ax1.plot(
    kh[cond], 1e-2 * kh[cond] ** (-5.0 / 3) * coef_norm[cond], "k-.", linewidth=1
)
plt.figtext(0.5, 0.45, "$k^{-5/3}$", fontsize=20)


ax1.set_xlabel("$k_h$")
ax1.set_ylabel("2D spectra")


plt.rc("legend", numpoints=1)
leg1 = plt.figlegend(
    [l_Etot[0], l_EK[0], l_EA[0]],
    ["$E$", "$E_K$", "$E_A$"],
    loc=(0.78, 0.7),
    labelspacing=0.2,
)


ax1.set_xlim([0.1, 150])
ax1.set_ylim([1e-5, 2e1])


create_fig.save_fig()

plt.show()



"""Forcing schemes (:mod:`fluidsim.base.forcing`)
=======================================================


Provides:

.. autosummary::
   :toctree:

   base
   specific
   anisotropic

"""

from .base import ForcingBase, ForcingBasePseudoSpectral



"""
Spatiotemporal Spectra (:mod:`fluidsim.solvers.ns2d.output.spatiotemporal_spectra`)
===================================================================================

Provides:

.. autoclass:: SpatioTemporalSpectraNS2D
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.output.spatiotemporal_spectra import (
    SpatioTemporalSpectra2D,
    SpatioTemporalSpectraNS,
)

from transonic import boost, Array, Type

A3 = Array[Type(np.float32, np.float64), "3d", "C"]
A2 = Array[np.float64, "2d", "C"]
A1 = "float[:]"


@boost
def compute_spectrum_kzkhomega(
    field_k0k1omega: A3, khs: A1, kzs: A1, KX: A2, KZ: A2, KH: A2
):
    """Compute the kz-kh-omega spectrum."""
    deltakh = khs[1]
    deltakz = kzs[1]

    nkh = len(khs)
    nkz = len(kzs)
    nk0, nk1, nomega = field_k0k1omega.shape
    spectrum_kzkhomega = np.zeros((nkz, nkh, nomega), dtype=field_k0k1omega.dtype)

    for ik0 in range(nk0):
        for ik1 in range(nk1):
            values = field_k0k1omega[ik0, ik1, :]
            kx = KX[ik0, ik1]
            if kx != 0.0:
                # warning: we should also consider another condition
                # (kx != kx_max) but it is not necessary here mainly
                # because of dealiasing
                values = 2 * values

            kappa = KH[ik0, ik1]
            ikh = int(kappa / deltakh)
            kz = abs(KZ[ik0, ik1])
            ikz = int(round(kz / deltakz))
            if ikz >= nkz - 1:
                ikz = nkz - 1
            if ikh >= nkh - 1:
                ikh = nkh - 1
                for i, value in enumerate(values):
                    spectrum_kzkhomega[ikz, ikh, i] += value
            else:
                coef_share = (kappa - khs[ikh]) / deltakh
                for i, value in enumerate(values):
                    spectrum_kzkhomega[ikz, ikh, i] += (1 - coef_share) * value
                    spectrum_kzkhomega[ikz, ikh + 1, i] += coef_share * value

    # get one-sided spectrum in the omega dimension
    nomega = (nomega + 1) // 2
    spectrum_onesided = np.zeros((nkz, nkh, nomega))
    spectrum_onesided[:, :, 0] = spectrum_kzkhomega[:, :, 0]
    spectrum_onesided[:, :, 1:] = (
        spectrum_kzkhomega[:, :, 1:nomega]
        + spectrum_kzkhomega[:, :, -1:-nomega:-1]
    )
    return spectrum_onesided / (deltakz * deltakh)


def _sum_wavenumber2D(field, KX, kx_max):
    n0, n1 = field.shape[:2]
    result = 0.0
    for i0 in range(n0):
        for i1 in range(n1):
            value = field[i0, i1]
            kx = KX[i0, i1]
            if kx != 0.0 and kx != kx_max:
                value *= 2
            result += value
    return result


class SpatioTemporalSpectraNS2D(SpatioTemporalSpectraNS, SpatioTemporalSpectra2D):

    compute_spectrum_kzkhomega = staticmethod(compute_spectrum_kzkhomega)
    _sum_wavenumber = staticmethod(_sum_wavenumber2D)

    def save_spectra_kzkhomega(self, tmin=0, tmax=None, dtype=None):
        return super().save_spectra_kzkhomega(tmin, tmax, dtype)



import unittest

from fluidsim.base.params import merge_params

from fluidsim.util.testing import TestCase

from fluidsim.solvers.ad1d.test_solver import scipy_installed


@unittest.skipIf(not scipy_installed, "No module named scipy.sparse")
class TestParameters(TestCase):
    """Test Parameters class and related functions."""

    def test_merge(self):
        """Test merging parameters."""
        from fluidsim.base.solvers.base import SimulBase
        from fluidsim.solvers.ad1d.solver import Simul

        params1 = SimulBase.create_default_params()
        params2 = Simul.create_default_params()
        merge_params(params1, params2)


if __name__ == "__main__":
    unittest.main()



"""State of the variables (:mod:`fluidsim.base.sphericalharmo.state`)
=====================================================================

Provides:

.. autoclass:: StateSphericalHarmo
   :members:
   :private-members:

"""

# from ..setofvariables import SetOfVariables
from ..state import StatePseudoSpectral


class StateSphericalHarmo(StatePseudoSpectral):
    """Contains the state variables and handles the access to fields.

    This is the general class for the pseudo-spectral solvers base on the
    Sperical harmonic transform.

    .. warning::

       We assume incompressibility (div = 0) in this base class!

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Static method to complete the ParamContainer info_solver."""

        StatePseudoSpectral._complete_info_solver(info_solver)

        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["rot_sh"],
                "keys_state_phys": ["ux", "uy", "rot"],
                "keys_computable": [],
                "keys_phys_needed": ["rot"],
                "keys_linear_eigenmodes": ["rot_sh"],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        """Compute and return a variable"""
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        results = {}

        if key == "ux_sh":
            result = self.oper.sht(self.state_phys.get_var("ux"))
        elif key == "uy_sh":
            result = self.oper.sht(self.state_phys.get_var("uy"))
        elif key == "div_sh":
            result = self.oper.create_array_sh(0.0)
        elif key == "div":
            result = self.oper.create_array_spat(0.0)
        elif key == "q":
            rot = self.state_phys.get_var("rot")
            result = rot
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

            for key, var in results.items():
                self.vars_computed[key] = var
                self.it_computed[key] = it

        return result

    def statephys_from_statespect(self):
        """Compute `state_phys` from `statespect`."""
        rot_sh = self.state_spect.get_var("rot_sh")

        # efficient
        rot = self.state_phys.get_var("rot")
        self.oper.isht_as_arg(rot_sh, rot)

        # not very efficient!
        ux, uy = self.oper.vec_from_rotsh(rot_sh)
        self.state_phys.set_var("ux", ux)
        self.state_phys.set_var("uy", uy)

    def statespect_from_statephys(self):
        """Compute `state_spect` from `state_phys`."""

        rot = self.state_phys.get_var("rot")
        rot_sh = self.state_spect.get_var("rot_sh")
        self.oper.sht_as_arg(rot, rot_sh)

    def init_from_rotsh(self, rot_sh):
        """Initialize the state from the variable `rot_sh`."""
        self.oper.dealiasing(rot_sh)
        self.state_spect.set_var("rot_sh", rot_sh)
        self.statephys_from_statespect()

    def init_statespect_from(self, **kwargs):
        if len(kwargs) == 1:
            if "rot_sh" in kwargs:
                self.init_from_rotsh(kwargs["rot_sh"])
        else:
            super().init_statespect_from(**kwargs)



import numpy as np

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.time_stepping.pseudo_spect import TimeSteppingPseudoSpectral

from .state import StateNS2D as StateBase
from .solver import InfoSolverNS2D as InfoBase, Simul as SimulBase, compute_Frot
from .output import Output as OutputBase


class TimeStepping(TimeSteppingPseudoSpectral):
    def one_time_step_computation(self):
        """One time step"""
        state_spect = self.sim.state.state_spect
        self._time_step_RK()
        self.sim.oper.dealiasing(state_spect)
        ux_fft = state_spect.get_var("ux_fft")
        uy_fft = state_spect.get_var("uy_fft")
        self.sim.oper.projection_perp(ux_fft, uy_fft)
        self.sim.state.statephys_from_statespect()
        if np.isnan(np.sum(state_spect[0])):
            raise ValueError(f"nan at it = {self.it}, t = {self.t:.4f}")


class State(StateBase):
    @staticmethod
    def _complete_info_solver(info_solver):
        StateBase._complete_info_solver(info_solver)
        info_solver.classes.State.keys_state_spect = ["ux_fft", "uy_fft"]

    def init_statespect_from(self, **kwargs):
        if len(kwargs) == 2 and sorted(kwargs.keys()) == ["ux_fft", "uy_fft"]:
            self.state_spect.set_var("ux_fft", kwargs["ux_fft"])
            self.state_spect.set_var("uy_fft", kwargs["uy_fft"])
            self.statephys_from_statespect()
            return
        super().init_statespect_from(**kwargs)

    def init_from_rotfft(self, rot_fft):
        ux_fft, uy_fft = self.sim.oper.vecfft_from_rotfft(rot_fft)
        self.state_spect.set_var("ux_fft", ux_fft)
        self.state_spect.set_var("uy_fft", uy_fft)
        self.statephys_from_statespect()

    def statephys_from_statespect(self):
        ux_fft = self.state_spect.get_var("ux_fft")
        uy_fft = self.state_spect.get_var("uy_fft")
        rot_fft = self.sim.oper.rotfft_from_vecfft(ux_fft, uy_fft)

        rot = self.state_phys.get_var("rot")
        ux = self.state_phys.get_var("ux")
        uy = self.state_phys.get_var("uy")

        self.oper.ifft_as_arg(rot_fft, rot)
        self.oper.ifft_as_arg(ux_fft, ux)
        self.oper.ifft_as_arg(uy_fft, uy)


class Output(OutputBase):
    def compute_energy_fft(self):
        """Compute energy(k)"""
        ux_fft = self.sim.state.state_spect.get_var("ux_fft")
        uy_fft = self.sim.state.state_spect.get_var("uy_fft")
        return (np.abs(ux_fft) ** 2 + np.abs(uy_fft) ** 2) / 2.0


class InfoSolver(InfoBase):
    def _init_root(self):

        super()._init_root()

        module = "fluidsim.solvers.ns2d.with_uxuy"
        self.module_name = module
        self.class_name = "Simul"
        self.short_name = "NS2D"

        classes = self.classes

        classes.State.module_name = module
        classes.State.class_name = "State"

        classes.Output.module_name = module
        classes.Output.class_name = "Output"

        classes.TimeStepping.module_name = module
        classes.TimeStepping.class_name = "TimeStepping"


class Simul(SimulBase):
    InfoSolver = InfoSolver

    def tendencies_nonlin(self, state_spect=None, old=None):

        # the operator and the fast Fourier transform
        oper = self.oper
        ifft_as_arg = oper.ifft_as_arg

        # get or compute rot_fft, ux and uy
        if state_spect is None:
            ux_fft = self.state.state_spect.get_var("ux_fft")
            uy_fft = self.state.state_spect.get_var("uy_fft")
            rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
            ux = self.state.state_phys.get_var("ux")
            uy = self.state.state_phys.get_var("uy")
        else:
            ux_fft = state_spect.get_var("ux_fft")
            uy_fft = state_spect.get_var("uy_fft")
            rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
            ux = self.state.field_tmp0
            uy = self.state.field_tmp1
            ifft_as_arg(ux_fft, ux)
            ifft_as_arg(uy_fft, uy)

        # "px" like $\partial_x$
        px_rot_fft, py_rot_fft = oper.gradfft_from_fft(rot_fft)

        px_rot = self.state.field_tmp2
        py_rot = self.state.field_tmp3

        ifft_as_arg(px_rot_fft, px_rot)
        ifft_as_arg(py_rot_fft, py_rot)

        Frot = compute_Frot(ux, uy, px_rot, py_rot, self.params.beta)

        if old is None:
            tendencies_fft = SetOfVariables(like=self.state.state_spect)
        else:
            tendencies_fft = old

        Frot_fft = oper.fft(Frot)
        oper.dealiasing(Frot_fft)

        Fx_fft, Fy_fft = self.oper.vecfft_from_rotfft(Frot_fft)

        tendencies_fft.set_var("ux_fft", Fx_fft)
        tendencies_fft.set_var("uy_fft", Fy_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft



import unittest

import numpy as np
import matplotlib.pyplot as plt

import fluidsim as fls

import fluiddyn.util.mpi as mpi

from fluidsim.solvers.ns2d.strat.solver import Simul
from fluidsim.solvers.ns2d.test_solver import TestSimulBase as Base


class TestSimulBase(Base):
    Simul = Simul


class TestSolverNS2DTendency(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.time_stepping.USE_CFL = False
        params.time_stepping.USE_T_END = False
        # todo: understand why it fails for larger it_end and deltat0. I (pa)
        # guess there could be a bug hidden.
        params.time_stepping.it_end = 2
        params.time_stepping.deltat0 = 0.02
        params.output.HAS_TO_SAVE = False

    def test_tendency(self):
        sim = self.sim

        rot_fft = sim.state.get_var("rot_fft")
        b_fft = sim.state.get_var("b_fft")

        assert not np.any(b_fft)

        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)
        Frot_fft = tend.get_var("rot_fft")
        Fb_fft = tend.get_var("b_fft")

        T_rot = np.real(Frot_fft.conj() * rot_fft)

        ratio = sim.oper.sum_wavenumbers(T_rot) / sim.oper.sum_wavenumbers(
            abs(T_rot)
        )

        self.assertGreater(1e-15, ratio)

        # init_fields = noise gives energy only to rot
        sim.time_stepping.start()

        rot_fft = sim.state.get_var("rot_fft")
        b_fft = sim.state.get_var("b_fft")

        # init_fields = noise gives energy only to rot
        assert np.any(b_fft)

        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)
        Frot_fft = tend.get_var("rot_fft")
        Fb_fft = tend.get_var("b_fft")

        assert sim.check_energy_conservation(rot_fft, b_fft, Frot_fft, Fb_fft)


class TestForcingLinearMode(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.forcing.enable = True
        params.forcing.type = "tcrandom_anisotropic"
        params.forcing.nkmin_forcing = 4
        params.forcing.nkmax_forcing = 6
        params.forcing.key_forced = "ap_fft"

    def test_forcing_linear_mode(self):
        sim = self.sim
        sim.time_stepping.start()
        if mpi.nb_proc == 1:
            sim.forcing.forcing_maker.plot_forcing_region()


class TestForcingConstantRateEnergy(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "tcrandom"
        params.forcing.normalized.constant_rate_of = "energy"
        params.forcing.key_forced = "rot_fft"
        params.forcing.tcrandom_anisotropic.kz_negative_enable = True
        params.forcing.forcing_rate = 3.333

        params.output.periods_save.spatial_means = 1e-6

        return params

    def test_(self):
        self.sim.time_stepping.start()

        if mpi.rank == 0:
            # Does the energy injection rate have the correct value at all times ?
            means = self.sim.output.spatial_means.load()
            P_tot = means["PK_tot"] + means["PA_tot"]
            assert np.allclose(P_tot, self.sim.params.forcing.forcing_rate)


class TestForcingConstantRateEnergyAP(TestForcingConstantRateEnergy):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.key_forced = "ap_fft"
        params.forcing.tcrandom_anisotropic.kz_negative_enable = False


class TestForcingOutput(TestSimulBase):
    @classmethod
    def init_params(self):

        params = super().init_params()

        # Time stepping parameters
        params.time_stepping.USE_CFL = False
        params.time_stepping.USE_T_END = False
        params.time_stepping.it_end = 2
        params.time_stepping.deltat0 = 0.1

        params.forcing.enable = True
        # params.forcing.type = "tcrandom"
        # Forcing also linear mode!!!
        params.forcing.type = "tcrandom_anisotropic"
        params.forcing.nkmin_forcing = 4
        params.forcing.nkmax_forcing = 6

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.1

        params.output.ONLINE_PLOT_OK = True
        params.output.periods_print.print_stdout = 0.2
        params.output.periods_plot.phys_fields = 0.2

        Lx, Ly = params.oper.Lx, params.oper.Ly
        nx, ny = params.oper.nx, params.oper.ny
        params.output.temporal_spectra.probes_region = (0.0, Lx, 0.0, Ly)
        params.output.temporal_spectra.probes_deltax = Lx / nx
        params.output.temporal_spectra.probes_deltay = Ly / ny
        params.output.temporal_spectra.SAVE_AS_FLOAT32 = True

        params.output.spatiotemporal_spectra.probes_region = (
            nx // 2 + 1,
            ny // 2 + 1,
        )
        params.output.spatiotemporal_spectra.SAVE_AS_COMPLEX64 = False

        for tag in params.output._tag_children:
            if tag.startswith("periods"):
                continue
            child = params.output[tag]
            if hasattr(child, "HAS_TO_PLOT_SAVED"):
                child["HAS_TO_PLOT_SAVED"] = True

    def test_forcing_output(self):

        sim = self.sim

        sim.time_stepping.start()

        sim.state.compute("rot_fft")
        sim.state.compute("rot_fft")

        with self.assertRaises(ValueError):
            sim.state.compute("abc")
        sim.state.compute("abc", RAISE_ERROR=False)

        ux_fft = sim.state.compute("ux_fft")
        uy_fft = sim.state.compute("uy_fft")

        sim.state.init_statespect_from(ux_fft=ux_fft)
        sim.state.init_statespect_from(uy_fft=uy_fft)

        sim.output.compute_enstrophy()

        if mpi.nb_proc == 1:

            plt.close("all")

            sim.output.plot_summary()

            sim.output.spectra.plot2d()

            sim.output.spatial_means.plot_energy()
            sim.output.spatial_means.plot_dt_energy()
            sim.output.spatial_means.plot_energy_shear_modes()

            plt.close("all")
            sim.output.spatial_means.compute_time_means()
            sim.output.spatial_means.load_dataset()
            sim.output.spatial_means.time_first_saved()
            sim.output.spatial_means.time_last_saved()

            sim.output.spectra_multidim.plot()

            with self.assertRaises(ValueError):
                sim.state.get_var("test")

        # `compute('q')` two times for better coverage...
        sim.state.get_var("q")
        sim.state.get_var("q")
        sim.state.get_var("div")

        path_run = sim.output.path_run
        if mpi.nb_proc > 1:
            path_run = mpi.comm.bcast(path_run)

        sim3 = fls.load_state_phys_file(path_run, modif_save_params=False)
        sim3.params.time_stepping.t_end += 0.2
        sim3.time_stepping.start()

        if mpi.nb_proc > 1:
            plt.close("all")
            return

        sim3.output.phys_fields.animate(
            "ux",
            dt_frame_in_sec=1e-6,
            dt_equations=0.3,
            repeat=False,
            clim=(-1, 1),
            save_file=False,
            numfig=1,
        )

        sim2 = fls.load_sim_for_plot(sim.output.path_run)
        sim2.output

        sim2.output.spect_energy_budg.load_mean()

        means = sim2.output.spatial_means.load()
        # `:-1` because the last time is saved twice in spatial_means
        # (see SpatialMeansBase.__init__)
        energy_K_mean = means["EK"][:-1].mean()

        sim2.output.temporal_spectra.load_time_series()
        spectra_omega = sim2.output.temporal_spectra.compute_spectra()
        omegas = spectra_omega["omegas"]
        delta_omega = omegas[1]

        sim2.output.temporal_spectra.plot_spectra()

        spatiotemporal_spectra = sim2.output.spatiotemporal_spectra
        series_kxky = spatiotemporal_spectra.load_time_series()

        spectra_kxkyomega = spatiotemporal_spectra.compute_spectra()
        spectra_omega_from_spatiotemp = (
            spatiotemporal_spectra.compute_temporal_spectra()
        )

        deltakx = 2 * np.pi / self.params.oper.Lx
        order = spectra_kxkyomega["dims_order"]
        KX = deltakx * spectra_kxkyomega[f"K{order[1]}_adim"]
        kx_max = self.params.oper.nx // 2 * deltakx

        assert kx_max == KX.max()

        from fluidsim.solvers.ns2d.output.spatiotemporal_spectra import (
            _sum_wavenumber2D,
        )

        def sum_wavenumber(field):
            return _sum_wavenumber2D(field, KX, kx_max)

        tspectrum_mean = (
            spectra_omega["spectrum_ux"] + spectra_omega["spectrum_uy"]
        )

        energy_tspect_mean = 0.5 * delta_omega * tspectrum_mean.sum()

        assert np.allclose(energy_K_mean, energy_tspect_mean), (
            energy_K_mean / energy_tspect_mean
        )

        _ = spatiotemporal_spectra.save_spectra_kzkhomega()
        spectra_kzkhomega = spatiotemporal_spectra.load_spectra_kzkhomega()

        delta_kz = spectra_kzkhomega["kz_spectra"][1]
        delta_kh = spectra_kzkhomega["kh_spectra"][1]
        delta_omega = spectra_kzkhomega["omegas"][1]
        coef = delta_kz * delta_kh * delta_omega

        for letter in "xy":
            vi_fft = series_kxky[f"u{letter}_Fourier"]
            spectrum_kxkykzomega = spectra_kxkyomega["spectrum_u" + letter]
            spectrum_omega = spectra_omega["spectrum_u" + letter]
            spectrum_omega_from_spatiotemp = spectra_omega_from_spatiotemp[
                "spectrum_u" + letter
            ]
            spectrum_kzkhomega = spectra_kzkhomega["spectrum_u" + letter]

            E_series_kxkykz = 0.5 * sum_wavenumber(
                (abs(vi_fft) ** 2).mean(axis=-1)
            )
            assert E_series_kxkykz > 0, (letter, vi_fft)

            E_kxkykzomega = (
                0.5
                * delta_omega
                * sum_wavenumber(spectrum_kxkykzomega.sum(axis=-1))
            )
            E_omega = 0.5 * delta_omega * spectrum_omega.sum()
            E_omega_from_spatiotemp = (
                0.5 * delta_omega * spectrum_omega_from_spatiotemp.sum()
            )

            E_kzkhomega = 0.5 * coef * spectrum_kzkhomega.sum()

            assert np.allclose(E_omega, E_kxkykzomega), (
                letter,
                E_kxkykzomega / E_omega,
            )

            assert np.allclose(E_series_kxkykz, E_kxkykzomega), (
                letter,
                E_kxkykzomega / E_series_kxkykz,
            )

            assert np.allclose(E_series_kxkykz, E_kzkhomega), (
                letter,
                E_kzkhomega / E_series_kxkykz,
            )

            assert np.allclose(E_omega, E_omega_from_spatiotemp), (
                letter,
                E_omega,
                E_omega_from_spatiotemp,
            )

        sim2.output.spatiotemporal_spectra.plot_kzkhomega(key_field="ux")
        sim2.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="ux", equation="omega=N/2"
        )
        sim2.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="ux", equation="kh=1"
        )
        sim2.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="ux", equation="kz=1"
        )
        sim2.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="ux", equation="ikh=1"
        )
        sim2.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="ux", equation="ikz=1"
        )
        sim2.output.spatiotemporal_spectra.plot_temporal_spectra()

        sim2.output.increments.load()
        sim2.output.increments.plot()
        sim2.output.increments.load_pdf_from_file()

        sim2.output.phys_fields.animate(
            "ux",
            dt_frame_in_sec=1e-6,
            dt_equations=0.3,
            repeat=False,
            clim=(-1, 1),
            save_file=False,
            numfig=1,
        )
        sim2.output.phys_fields.plot()

        plt.close("all")


class TestSolverNS2DInitLinearMode(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.init_fields.type = "linear_mode"
        params.output.HAS_TO_SAVE = False

    def test_init_linear_mode(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""
Module where the version is written.

It is executed in setup.py and imported in fluidsim/__init__.py.

See:

http://en.wikipedia.org/wiki/Software_versioning
http://legacy.python.org/dev/peps/pep-0386/

'a' or 'alpha' means alpha version (internal testing),
'b' or 'beta' means beta version (external testing).

PEP 440 also permits the use of local version identifiers. This is initialized
using the setuptools_scm module, if available.

See:
https://www.python.org/dev/peps/pep-0440/#local-version-identifiers
https://github.com/pypa/setuptools_scm#setuptools_scm

"""

# NOTE: Version autogenerated by setup.py. Do not manually edit here.
# Update version in lib/fluidsim_core/_version.py instead.
__version__ = "0.5.0.post0"


__all__ = ["__version__", "get_local_version", "__about__"]

try:
    from pyfiglet import figlet_format

    __about__ = figlet_format("fluidsim", font="big")
except ImportError:
    __about__ = r"""
  __ _       _     _     _
 / _| |     (_)   | |   (_)
| |_| |_   _ _  __| |___ _ _ __ ___
|  _| | | | | |/ _` / __| | '_ ` _ \
| | | | |_| | | (_| \__ \ | | | | | |
|_| |_|\__,_|_|\__,_|___/_|_| |_| |_|
"""

__about__ = __about__.rstrip() + f"\n\n{28 * ' '} v. {__version__}\n"

_loc_version = None


def get_local_version():
    """Get a long "local" version."""

    global _loc_version

    if _loc_version is None:
        from setuptools_scm import get_version

        try:
            _loc_version = get_version(root="..", relative_to=__file__)
        except (LookupError, AssertionError):
            _loc_version = __version__

    return _loc_version



"""Output (:mod:`fluidsim.base.output`)
=============================================

Provides:

.. autosummary::
   :toctree:

   base
   prob_dens_func
   spectra
   phys_fields
   phys_fields1d
   phys_fields2d
   phys_fields3d
   movies
   spatial_means
   time_signals_fft
   increments
   print_stdout
   spect_energy_budget


.. autoclass:: OutputBase
   :members:
   :private-members:

.. autoclass:: OutputBasePseudoSpectral
   :members:
   :private-members:

"""

import os
from glob import glob
import argparse

import h5py
import matplotlib as mpl

import fluiddyn.output

from .base import OutputBase, OutputBasePseudoSpectral

mpl.rc("axes", titlesize=10)

__all__ = ["OutputBase", "OutputBasePseudoSpectral"]


def create_description_xmf_files(path=None):
    """Create description xmf files for Paraview"""

    raise DeprecationWarning(
        "create_description_xmf_files is deprecated. "
        "Use create_description_xmf_file instead."
    )


def create_description_xmf_file(path=None):
    """Create description xmf file for Paraview"""

    if path is None:
        path = os.getcwd()

    if os.path.isdir(path):
        paths = glob(path + "/state_phys*.[hn][5c]")
        path_dir = path
    else:
        paths = glob(path)
        path_dir = os.path.dirname(path)

    if len(paths) == 0:
        raise ValueError(f"No file corresponds to this path {path}.")

    path_out = os.path.join(path_dir, "states_phys.xmf")

    paths.sort()
    path = paths[0]

    with h5py.File(path, "r") as file:
        ndim = 3

        nx = file["/info_simul/params/oper"].attrs["nx"]
        Lx = file["/info_simul/params/oper"].attrs["Lx"]
        deltax = Lx / nx
        try:
            ny = file["/info_simul/params/oper"].attrs["ny"]
            Ly = file["/info_simul/params/oper"].attrs["Ly"]
            deltay = Ly / ny
        except KeyError:
            ndim = 1
        try:
            nz = file["/info_simul/params/oper"].attrs["nz"]
            Lz = file["/info_simul/params/oper"].attrs["Lz"]
            deltaz = Lz / nz
        except KeyError:
            ndim = 2

        keys = list(file["/state_phys"].keys())

    if ndim == 1:
        geometry_type = "Origin_Dx"
        dims_data = f"{nx}"
        origins = "0"
        deltaxs = f"{deltax}"
    elif ndim == 2:
        geometry_type = "Origin_DxDy"
        dims_data = f"{ny} {nx}"
        origins = "0 0"
        deltaxs = f"{deltay} {deltax}"
    elif ndim == 3:
        geometry_type = "Origin_DxDyDz"
        dims_data = f"{nz} {ny} {nx}"
        origins = "0 0 0"
        deltaxs = f"{deltaz} {deltay} {deltax}"

    if ndim in (2, 3):
        vectors = []
        if ndim == 2:
            components = ("x", "y")
            for_join = "$0, $1"
        elif ndim == 3:
            components = ("x", "y", "z")
            for_join = "$0, $1, $2"

        for key in keys:
            if key.endswith("x"):
                vector = key[:-1]
                vector_components = {vector + compo for compo in components}
                if vector_components.issubset(set(keys)):
                    vectors.append(vector)

    txt = """<?xml version="1.0" ?>
<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
<Xdmf>
  <Domain>
    <Grid Name="TimeSeries" GridType="Collection"
          CollectionType="Temporal">
    """

    for path in paths:
        base_name = os.path.basename(path)

        with h5py.File(path, "r") as file:
            time = file["state_phys"].attrs["time"]

        txt += """
    <Grid Name="my_uniform_grid" GridType="Uniform">
      <Time Value="{time:.9f}" />
      <Topology TopologyType="{ndim}DCoRectMesh" Dimensions="{dims_data}">
      </Topology>
      <Geometry GeometryType="{geometry_type}">
        <DataItem Dimensions="{ndim}" NumberType="Float" Format="XML">
        {origins}
        </DataItem>
        <DataItem Dimensions="{ndim}" NumberType="Float" Format="XML">
        {deltaxs}
        </DataItem>
      </Geometry>
""".format(
            time=float(time),
            ndim=ndim,
            geometry_type=geometry_type,
            dims_data=dims_data,
            origins=origins,
            deltaxs=deltaxs,
        )

        for key in keys:
            txt += """
      <Attribute Name="{key}" AttributeType="Scalar" Center="Node">
        <DataItem Dimensions="{dims_data}" NumberType="Float" Format="HDF">
          {file_name}:/state_phys/{key}
        </DataItem>
      </Attribute>
""".format(
                key=key, dims_data=dims_data, file_name=base_name
            )

        for vector in vectors:
            txt += """
      <Attribute Name="{vector}" AttributeType="Vector" Center="Node">
        <DataItem Dimensions="{dims_data} {ndim}"  ItemType="Function"
                  Function="JOIN({for_join})">
          <DataItem Dimensions="{dims_data}" NumberType="Float" Format="HDF">
            {file_name}:/state_phys/{vector}x
          </DataItem>
          <DataItem Dimensions="{dims_data}" NumberType="Float" Format="HDF">
            {file_name}:/state_phys/{vector}y
          </DataItem>
          <DataItem Dimensions="{dims_data}" NumberType="Float" Format="HDF">
            {file_name}:/state_phys/{vector}z
          </DataItem>
        </DataItem>
      </Attribute>
""".format(
                vector=vector,
                dims_data=dims_data,
                for_join=for_join,
                ndim=ndim,
                file_name=base_name,
            )

        txt += "    </Grid>"

    txt += """
    </Grid>
  </Domain>
</Xdmf>
"""

    with open(path_out, "w") as file:
        file.write(txt)

    print(
        "Creation of the file "
        + path_out
        + "\nOpen it with a Xdmf reader to read the output files."
    )


def run():

    parser = argparse.ArgumentParser(
        prog="fluidsim-create-xml-description",
        description=create_description_xmf_file.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "str_files",
        nargs="?",
        default=os.getcwd(),
        help="str indicating which file has to be dump.",
        type=str,
    )

    args = parser.parse_args()

    create_description_xmf_file(args.str_files)



import numpy as np

from fluiddyn.calcul.easypyfft import FFTW1DReal2Complex, fftw_grid_size


class FrequencyModulatedSignalMaker:
    def __init__(self, total_time, approximate_dt):
        self.nb_points = 2 * fftw_grid_size(int(total_time / approximate_dt))
        self.dt = total_time / self.nb_points
        self.times = self.dt * np.arange(self.nb_points)

        self.oper_fft = FFTW1DReal2Complex(self.nb_points)
        nb_omegas = self.oper_fft.shapeK[0]
        domega = 2 * np.pi / total_time
        self.omegas = domega * np.arange(nb_omegas)
        self.omegas_nozero = self.omegas.copy()
        self.omegas_nozero[0] = 1e-14

    def create_frequency_modulated_signal(
        self,
        omega_f,
        delta_omega_f,
        amplitude=1.0,
        omega_min=None,
        omega_max=None,
    ):
        # filter parameters for the time signal
        if omega_min is None:
            omega_min = omega_f / 10
        if omega_max is None:
            omega_max = omega_f
        # Gaussian white noise
        omegaf_vs_time = np.random.randn(self.nb_points)
        omegaf_vs_omega = self.oper_fft.fft(omegaf_vs_time)
        # rectangular filtering
        omegaf_vs_omega *= self.omegas >= omega_min
        omegaf_vs_omega *= self.omegas <= omega_max
        # set amplitude
        omegaf_vs_time = self.oper_fft.ifft(omegaf_vs_omega)
        omegaf_vs_time *= delta_omega_f / np.std(omegaf_vs_time)
        # time integration for frequency modulation
        omegaf_vs_omega = self.oper_fft.fft(omegaf_vs_time)
        phase_vs_omega = omegaf_vs_omega / (1j * self.omegas_nozero)
        phase_vs_time = self.oper_fft.ifft(phase_vs_omega)
        # frequency modulated forcing time signal
        phase_vs_time += omega_f * self.times - phase_vs_time[0]
        forcing_vs_time = (
            amplitude * (omegaf_vs_time + omega_f) * np.sin(phase_vs_time)
        )
        return self.times, forcing_vs_time



"""SW1L with linear terms solved exactly (:mod:`fluidsim.solvers.sw1l.exactlin`)
================================================================================

.. autosummary::
   :toctree:

   solver
   state

.. autosummary::
   :toctree:

   modified

"""



"""Multidimensional spectra output (:mod:`fluidsim.solvers.ns2d.strat.output.spectra_multidim`)
===============================================================================================

.. autoclass:: SpectraMultiDimNS2DStrat
   :members:
   :private-members:

"""

import h5py
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from math import radians
from fluidsim.base.output.spectra_multidim import SpectraMultiDim


class SpectraMultiDimNS2DStrat(SpectraMultiDim):
    """Save and plot the spectra."""

    def compute(self):
        """Computes multidimensional spectra at one time."""

        # Get variables
        energyK_fft, energyA_fft = self.output.compute_energies_fft()

        ap_fft = self.sim.state.compute("ap_fft")
        am_fft = self.sim.state.compute("am_fft")

        # Computes multidimensional spectra
        spectrumkykx_EK = self.oper.compute_spectrum_kykx(
            energyK_fft, folded=False
        )
        spectrumkykx_EA = self.oper.compute_spectrum_kykx(
            energyA_fft, folded=False
        )

        energy_ap_fft = abs(ap_fft) ** 2
        energy_am_fft = abs(am_fft) ** 2

        spectrumkykx_ap = self.oper.compute_spectrum_kykx(
            energy_ap_fft, folded=False
        )
        spectrumkykx_am = self.oper.compute_spectrum_kykx(
            energy_am_fft, folded=False
        )

        dict_spectra = {
            "spectrumkykx_EK": spectrumkykx_EK,
            "spectrumkykx_EA": spectrumkykx_EA,
            "spectrumkykx_ap": spectrumkykx_ap,
            "spectrumkykx_am": spectrumkykx_am,
        }

        return dict_spectra

    # def _online_plot_saving(self, dict_spectra):
    #     pass

    def plot(self, key=None, tmin=None, tmax=None, xlim=None, zlim=None):
        """
        Plots spectrumkykx averaged between tmin and tmax.

        Parameters
        ----------
        key : str
          Key to plot the spectrum: E, EK, EA, ap_fft (default), am_fft

        tmin : float
          Lower time to compute the time average.

        tmax : float
          Upper time to compute the time average.

        xlim : float
          Upper limit kx to plot.

        zlim : float
          Upper limit kz to plot.

        """

        oper = self.sim.params.oper
        pforcing = self.sim.params.forcing

        # Load data
        with h5py.File(self.path_file, "r") as file:
            times = file["times"][...]
            kx = file["kxE"][...]
            ap_fft_spectrum = file["spectrumkykx_ap"]
            if key == "EK":
                data = file["spectrumkykx_EK"]
            elif key == "EA":
                data = file["spectrumkykx_EA"]
            elif key == "ap_fft" or not key:
                data = file["spectrumkykx_ap"]
                text_plot = r"$\hat{a}_+$"
            elif key == "am_fft":
                data = file["spectrumkykx_am"]
                text_plot = r"$\hat{a}_-$"
            else:
                raise ValueError("Key unknown.")

            # Compute time average
            if tmin is None:
                tmin = times[0]
            if tmax is None:
                tmax = times[-1]

            itmin = np.argmin(abs(times - tmin))
            if tmin == tmax:
                data_plot = data[itmin, :, :]
                if key != "ap_fft":
                    vmax_modified = ap_fft_spectrum[itmin, :, :].max()

            else:
                itmax = np.argmin(abs(times - tmax))
                data_plot = (data[itmin:itmax, :, :]).mean(0)
                if key != "ap_fft":
                    vmax_modified = (
                        (ap_fft_spectrum[itmin:itmax, :, :]).mean(0)
                    ).max()

        # Create array kz with negative values
        kz = 2 * np.pi * np.fft.fftfreq(oper.ny, oper.Ly / oper.ny)
        kz[kz.shape[0] // 2] *= -1

        # Create mesh of wave-numbers
        KX, KZ = np.meshgrid(kx, kz)

        # Set figure parameters
        fig, ax = plt.subplots()
        ax.set_xlabel(r"$k_x$", fontsize=14)
        ax.set_ylabel(r"$k_z$", fontsize=14)

        # Set axis limit
        if xlim:
            ikx = np.argmin(abs(kx - xlim))
            ax.set_xlim([0, kx[ikx] - self.sim.oper.deltakx])
        else:
            ikx = np.argmin(abs(kx - kx.max()))

        if zlim:
            ikz = np.argmin(abs(kz - zlim))
            ikz_negative = np.argmin(abs(kz + zlim))
            ax.set_ylim([kz[ikz_negative], kz[ikz] - self.sim.oper.deltaky])
        else:
            ikz = np.argmin(abs(kz - kz.max()))

        # Modify grid
        kz_modified = np.empty_like(kz)
        kz_modified[0 : kz_modified.shape[0] // 2 - 1] = kz[
            kz_modified.shape[0] // 2 + 1 :
        ]
        kz_modified[kz_modified.shape[0] // 2 - 1 :] = kz[
            0 : kz_modified.shape[0] // 2 + 1
        ]

        KX, KZ = np.meshgrid(kx, kz_modified)

        data_plot_modified = np.empty_like(data_plot)
        data_plot_modified[0 : kz_modified.shape[0] // 2 - 1, :] = data_plot[
            kz_modified.shape[0] // 2 + 1 :, :
        ]
        data_plot_modified[kz_modified.shape[0] // 2 - 1 :, :] = data_plot[
            0 : kz_modified.shape[0] // 2 + 1, :
        ]

        # Vmax is 1% of the maximum value.
        vmin = 0
        vmax = 0.01 * data_plot_modified.max()
        if key != "ap_fft":
            vmax = 0.01 * vmax_modified

        print("vmax", vmax)

        spectrum = ax.pcolormesh(
            KX, KZ, data_plot_modified, shading="nearest", vmin=vmin, vmax=vmax
        )

        # Create a Rectangle patch
        deltak = max(self.sim.oper.deltakx, self.sim.oper.deltaky)

        if isinstance(pforcing.tcrandom_anisotropic.angle, str):
            from math import radians

            angle = radians(
                float((pforcing.tcrandom_anisotropic.angle).split("°")[0])
            )
        else:
            angle = pforcing.tcrandom_anisotropic.angle

        x_rect = np.sin(angle) * deltak * pforcing.nkmin_forcing

        z_rect = np.cos(angle) * deltak * pforcing.nkmin_forcing

        width = abs(x_rect - np.sin(angle) * deltak * pforcing.nkmax_forcing)

        height = abs(z_rect - np.cos(angle) * deltak * pforcing.nkmax_forcing)

        rect1 = patches.Rectangle(
            (x_rect, z_rect),
            width,
            height,
            linewidth=1,
            edgecolor="r",
            facecolor="none",
        )

        ax.add_patch(rect1)

        if pforcing.tcrandom_anisotropic.kz_negative_enable:
            rect2 = patches.Rectangle(
                (x_rect, -(z_rect + height)),
                width,
                height,
                linewidth=1,
                edgecolor="r",
                facecolor="none",
            )

            ax.add_patch(rect2)

        # Plot arc kmin and kmax forcing
        ax.add_patch(
            patches.Arc(
                xy=(0, 0),
                width=2 * pforcing.nkmin_forcing * deltak,
                height=2 * pforcing.nkmin_forcing * deltak,
                angle=0,
                theta1=-90.0,
                theta2=90.0,
                linestyle="-.",
                color="red",
            )
        )
        ax.add_patch(
            patches.Arc(
                xy=(0, 0),
                width=2 * pforcing.nkmax_forcing * deltak,
                height=2 * pforcing.nkmax_forcing * deltak,
                angle=0,
                theta1=-90,
                theta2=90.0,
                linestyle="-.",
                color="red",
            )
        )

        # Text at 70% of the xlim and ylim
        ikx_text = np.argmin(abs(kx - kx[ikx] * 0.7))
        ikz_text = np.argmin(abs(kz - kz[ikz] * 0.7))

        ax.text(
            kx[ikx_text], kz[ikz_text], f"{text_plot}", color="white", fontsize=15
        )

        # Colorbar
        fig.colorbar(spectrum)

        ax.set_aspect("equal")

        return fig



"""NS2D Boussinesq solver (:mod:`fluidsim.solvers.ns2d.bouss.solver`)
=====================================================================

.. autoclass:: Simul
   :members:
   :private-members:

"""
import numpy as np

from transonic import jit, Array

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.solvers.ns2d.solver import InfoSolverNS2D, Simul as SimulNS2D


AF = Array[np.float64, "2d"]


@jit
def tendencies_nonlin_ns2dbouss(
    ux: AF, uy: AF, px_rot: AF, py_rot: AF, px_b: AF, py_b: AF
):
    Frot = -ux * px_rot - uy * py_rot + px_b
    Fb = -ux * px_b - uy * py_b
    return Frot, Fb


class InfoSolverNS2DBouss(InfoSolverNS2D):
    def _init_root(self):

        super()._init_root()

        package = "fluidsim.solvers.ns2d.bouss"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "ns2d.bouss"

        classes = self.classes

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateNS2DBouss"


# classes.InitFields.module_name = package + '.init_fields'
# classes.InitFields.class_name = 'InitFieldsNS2DStrat'

# classes.Output.module_name = package + '.output'
# classes.Output.class_name = 'OutputStrat'

# classes.Forcing.module_name = 'fluidsim.solvers.ns2d' + '.forcing'
# classes.Forcing.class_name = 'ForcingNS2D'

# classes.TimeStepping.module_name = \
#     package + '.time_stepping'

# classes.TimeStepping.class_name = 'TimeSteppingPseudoSpectralStrat'


class Simul(SimulNS2D):
    """Pseudo-spectral solver 2D incompressible Navier-Stokes equations."""

    InfoSolver = InfoSolverNS2DBouss

    def tendencies_nonlin(self, state_spect=None, old=None):
        r"""Compute the nonlinear tendencies of the solver ns2d.strat.

        Parameters
        ----------

        state_spect : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state, i.e. the vorticity, in Fourier
            space.  If `state_spect`, the variables vorticity and the
            velocity are computed from it, otherwise, they are taken
            from the global state of the simulation, `self.state`.

            These two possibilities are used during the Runge-Kutta
            time-stepping.

        Returns
        -------

        tendencies_fft : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the vorticity and the
            buoyancy.

        Notes
        -----

        .. |p| mathmacro:: \partial

        The 2D Navier-Stokes equation can be written

        .. math:: \p_t \hat\zeta = \hat N(\zeta) - \sigma(k) \hat \zeta,

        and

        .. math:: \p_t \hat b = \hat N(b) - \sigma(k) \hat b

        This function compute the nonlinear terms ("tendencies") :math:`\hat
        N(\zeta) = - \mathbf{u}\cdot \mathbf{\nabla} \zeta +
        \mathbf{\nabla}\wedge b\mathbf{e_z} = - \mathbf{u}\cdot \mathbf{\nabla}
        \zeta + \p_x b` and :math:`\hat N(b) = - \mathbf{u}\cdot
        \mathbf{\nabla} b`."""
        oper = self.oper
        fft_as_arg = oper.fft_as_arg
        ifft_as_arg = oper.ifft_as_arg

        if old is None:
            tendencies_fft = SetOfVariables(like=self.state.state_spect)
        else:
            tendencies_fft = old
        Frot_fft = tendencies_fft.get_var("rot_fft")
        Fb_fft = tendencies_fft.get_var("b_fft")

        if state_spect is None:
            rot_fft = self.state.state_spect.get_var("rot_fft")
            b_fft = self.state.state_spect.get_var("b_fft")
            ux = self.state.state_phys.get_var("ux")
            uy = self.state.state_phys.get_var("uy")
        else:
            rot_fft = state_spect.get_var("rot_fft")
            b_fft = state_spect.get_var("b_fft")
            ux_fft, uy_fft = oper.vecfft_from_rotfft(rot_fft)
            ux = self.state.field_tmp0
            uy = self.state.field_tmp1
            ifft_as_arg(ux_fft, ux)
            ifft_as_arg(uy_fft, uy)

        px_rot_fft, py_rot_fft = oper.gradfft_from_fft(rot_fft)
        px_b_fft, py_b_fft = oper.gradfft_from_fft(b_fft)

        px_rot = self.state.field_tmp2
        py_rot = self.state.field_tmp3

        px_b = self.state.field_tmp4
        py_b = self.state.field_tmp5

        ifft_as_arg(px_rot_fft, px_rot)
        ifft_as_arg(py_rot_fft, py_rot)
        ifft_as_arg(px_b_fft, px_b)
        ifft_as_arg(py_b_fft, py_b)

        Frot, Fb = tendencies_nonlin_ns2dbouss(ux, uy, px_rot, py_rot, px_b, py_b)

        fft_as_arg(Fb, Fb_fft)
        fft_as_arg(Frot, Frot_fft)

        oper.dealiasing(tendencies_fft)

        # # CHECK NON-LINEAR TRANSFER rot
        # T_rot = np.real(Frot_fft.conj()*rot_fft + Frot_fft*rot_fft.conj())/2.
        # print(('sum(T_rot) = {0:9.4e} ; sum(abs(T_rot)) = {1:9.4e}').format(
        #     self.oper.sum_wavenumbers(T_rot),
        #     self.oper.sum_wavenumbers(abs(T_rot))))

        # # CHECK NON-LINEAR TRANSFER b
        # T_b = np.real(Fb_fft.conj()*b_fft + Fb_fft*b_fft.conj())/2.
        # print(('sum(T_b) = {0:9.4e} ; sum(abs(T_b)) = {1:9.4e}').format(
        #     self.oper.sum_wavenumbers(T_b),
        #     self.oper.sum_wavenumbers(abs(T_b))))

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        # CHECK ENERGY CONSERVATION
        # Nrot_fft = tendencies_fft.get_var('rot_fft')
        # Nb_fft = tendencies_fft.get_var('b_fft')
        # self.check_energy_conservation(rot_fft, b_fft, Nrot_fft, Nb_fft)

        return tendencies_fft


# def check_energy_conservation(self, rot_fft, b_fft, N_rot, N_b):
#     """ Check energy conservation for the inviscid case. """
#     oper = self.oper
#     params = self.params

#     # Compute time derivative kinetic energy
#     division = 1./(oper.KX**2 + oper.KY**2)
#     division[np.where(division == np.inf)] = 0

#     pt_energyK_fft = 0.5 * division * np.real(rot_fft.conj() * N_rot)
#     pt_energyK_fft[np.isinf(pt_energyK_fft)] = 0.

#     # Compute time derivative potential energy
#     pt_energyA_fft = (1./(2 * params.N**2)) * np.real(b_fft.conj() * N_b)

#     # Time derivative total energy
#     pt_energy_fft = pt_energyK_fft + pt_energyA_fft

#     # Check time derivative energy is ~ 0.
#     epsilon = 1e-8
#     pt_energy = self.output.sum_wavenumbers(pt_energy_fft)
#     energy_conserved = np.abs(pt_energy) < epsilon
#     if not energy_conserved:
#         print('Energy conserved = ', energy_conserved)
#         print('pt_energy = {}'.format(
#             np.abs(self.output.sum_wavenumbers(pt_energy_fft))))

# def compute_dispersion_relation(self):
#     """
#     Computes the dispersion relation of internal gravity waves solver
#     ns2d.strat.

#     Returns
#     -------
#     omega_dispersion_relation : arr
#       Frequency dispersion relation in rad.
#     """
#     return self.params.N * (self.oper.KX / self.oper.K_not0)


if __name__ == "__main__":

    from math import pi

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    params.oper.nx = nx = 64
    params.oper.ny = nx // 2
    params.oper.Lx = Lx = 2 * pi
    params.oper.Ly = Lx / 2
    params.oper.coef_dealiasing = 0.5

    delta_x = Lx / nx

    params.nu_2 = 1.0 * 10e-6
    # params.nu_8 = 2.*10e-1*params.forcing.forcing_rate**(1./3)*delta_x**8
    params.time_stepping.USE_CFL = True
    params.time_stepping.USE_T_END = True
    # params.time_stepping.deltat0 = 0.1
    # Period of time of the simulation
    params.time_stepping.t_end = 5.0
    # params.time_stepping.it_end = 50

    params.init_fields.type = "noise"

    params.forcing.enable = True
    # params.forcing.type = 'tcrandom_anisotropic'
    params.forcing.type = "user_defined"

    # params.forcing.nkmax_forcing = 12
    # params.forcing.nkmin_forcing = 4
    # params.forcing.tcrandom_anisotropic.angle = '45'

    # 'Proportional'
    # params.forcing.type_normalize

    params.output.sub_directory = "tests"

    params.output.periods_print.print_stdout = 0.01

    params.output.periods_save.phys_fields = 2.0
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spatial_means = 0.05
    params.output.periods_save.spect_energy_budg = 1.0
    params.output.periods_save.increments = 1.0

    params.output.periods_plot.phys_fields = 5.0

    params.output.ONLINE_PLOT_OK = True

    params.output.spectra.HAS_TO_PLOT_SAVED = True
    params.output.spatial_means.HAS_TO_PLOT_SAVED = True
    params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = False
    params.output.increments.HAS_TO_PLOT_SAVED = False

    params.output.phys_fields.field_to_plot = "rot"

    sim = Simul(params)

    # monkey-patching for forcing
    import numpy as np

    forcing_maker = sim.forcing.forcing_maker
    oper = forcing_maker.oper_coarse
    forcing0 = np.cos(2 * np.pi * oper.Y / oper.ly)
    omega = 2 * np.pi

    def compute_forcingc_each_time(self):
        return forcing0 * np.sin(omega * sim.time_stepping.t)

    forcing_maker.monkeypatch_compute_forcingc_each_time(
        compute_forcingc_each_time
    )

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    sim.output.phys_fields.plot()

    fld.show()



"""Text output (:mod:`fluidsim.solvers.models0d.lorenz.output.print_stdout`)
============================================================================

.. autoclass:: PrintStdOutLorenz
   :members:
   :private-members:

"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from fluidsim.base.output.print_stdout import PrintStdOutBase

from fluiddyn.util import mpi


class PrintStdOutLorenz(PrintStdOutBase):
    """Simple text output.

    Used to print in both the stdout and the stdout.txt file, and also
    to print simple info on the current state of the simulation.

    """

    def complete_init_with_state(self):

        if self.period_print == 0:
            return

        self.t_last_print_info = -self.period_print

        self.print_stdout = self.__call__

    def _make_str_info(self):
        to_print = super()._make_str_info()

        if mpi.rank == 0:
            to_print += (
                (" " * 14) + "X = {:9.3e} ; Y = {:+9.3e} ; Z = {:+9.3e}\n" + "\n"
            ).format(
                float(self.sim.state.state_phys.get_var("X")),
                float(self.sim.state.state_phys.get_var("Y")),
                float(self.sim.state.state_phys.get_var("Z")),
            )

            duration_left = self._evaluate_duration_left()
            if duration_left is not None:
                to_print += f"              estimated remaining duration = {duration_left}"

        return to_print

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}
        with open(self.output.path_run + "/stdout.txt") as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_X = []
        for il, line in enumerate(lines):
            if line.startswith("it ="):
                lines_t.append(line)
            if line.startswith(" " * 14 + "X ="):
                lines_X.append(line)

        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        it = np.zeros(nt, dtype=int)
        t = np.zeros(nt)
        deltat = np.zeros(nt)

        X = np.zeros(nt)
        Y = np.zeros(nt)
        Z = np.zeros(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            it[il] = int(words[2])
            t[il] = float(words[6])
            deltat[il] = float(words[10])

            line = lines_X[il]
            words = line.split()
            X[il] = float(words[2])
            Y[il] = float(words[6])
            Z[il] = float(words[10])

        dict_results["it"] = it
        dict_results["t"] = t
        dict_results["deltat"] = deltat
        dict_results["X"] = X
        dict_results["Y"] = Y
        dict_results["Z"] = Z

        return dict_results

    def plot_deltat(self):
        dict_results = self.load()

        t = dict_results["t"]
        deltat = dict_results["deltat"]

        size_axe = [0.12, 0.12, 0.8, 0.8]
        fig, ax = self.output.figure_axe(size_axe=size_axe)

        ax.set_xlabel("t")
        ax.set_ylabel("deltat(t)")

        ax.set_title("info stdout\n" + self.output.summary_simul)
        ax.plot(t, deltat, "k", linewidth=2)

    def plot_XY_vs_time(self):
        dict_results = self.load()
        t = dict_results["t"]
        X = dict_results["X"]
        Y = dict_results["Y"]

        size_axe = [0.12, 0.12, 0.8, 0.8]
        fig, ax = self.output.figure_axe(size_axe=size_axe)

        ax.set_xlabel("$t$")
        ax.set_ylabel("$X$, $Y$")

        ax.plot(t, X, "b", label="$X$")
        ax.plot(t, Y, "r", label="$Y$")

        ax.plot(ax.get_xlim(), [self.sim.Xs0] * 2, "b--")
        ax.plot(ax.get_xlim(), [self.sim.Ys0] * 2, "r--")

        ax.plot(ax.get_xlim(), [self.sim.Xs1] * 2, "b.-")
        ax.plot(ax.get_xlim(), [self.sim.Ys1] * 2, "r.-")

        ax.legend()

    def plot_XY(self):
        dict_results = self.load()
        X = dict_results["X"]
        Y = dict_results["Y"]

        size_axe = [0.12, 0.12, 0.8, 0.8]
        fig, ax = self.output.figure_axe(size_axe=size_axe)

        ax.set_xlabel("$X$")
        ax.set_ylabel("$Y$")

        ax.plot(X, Y, "b")

        ax.plot(self.sim.Xs0, self.sim.Ys0, "bx")

        ax.plot(self.sim.Xs1, self.sim.Ys1, "bx")

    def plot_XZ(self):
        dict_results = self.load()
        X = dict_results["X"]
        Z = dict_results["Z"]

        size_axe = [0.12, 0.12, 0.8, 0.8]
        fig, ax = self.output.figure_axe(size_axe=size_axe)

        ax.set_xlabel("$X$")
        ax.set_ylabel("$Z$")

        ax.plot(X, Z, "b")

        ax.plot(self.sim.Xs0, self.sim.Zs0, "bx")

        ax.plot(self.sim.Xs1, self.sim.Zs1, "bx")

    def plot_XYZ(self):
        dict_results = self.load()
        X = dict_results["X"]
        Y = dict_results["Y"]
        Z = dict_results["Z"]

        fig = plt.figure()
        size_axe = [0.12, 0.12, 0.8, 0.8]

        ax = fig.add_axes(size_axe, projection="3d")

        ax.set_xlabel("$X$")
        ax.set_ylabel("$Y$")
        ax.set_zlabel("$Z$")

        ax.plot(X, Y, Z, "b")

        ax.plot([self.sim.Xs0], [self.sim.Ys0], self.sim.Zs0, "bx")
        ax.plot([self.sim.Xs1], [self.sim.Ys1], self.sim.Zs1, "bx")



from pathlib import Path
import unittest

import numpy as np
import h5py

from fluiddyn.util import mpi

from fluidsim.solvers.ns3d.test_solver import (
    TestSimulBase,
    skip_if_no_fluidfft,
    classproperty,
)

from fluidsim.util.util import (
    modif_resolution_from_dir,
    modif_resolution_from_dir_memory_efficient,
)


@unittest.skipIf(mpi.nb_proc > 1, "Modif resolution do not work with mpi")
@skip_if_no_fluidfft
class TestModifResol3d(TestSimulBase):
    def test_modif(self):
        self.sim.output.phys_fields.save()
        self.sim.output.close_files()

        path_run = self.sim.output.path_run

        # input parameters of the modif_resol functions
        t_approx = None
        coef_modif_resol = 3 / 2

        # first, the standard function
        modif_resolution_from_dir(
            path_run, t_approx, coef_modif_resol, PLOT=False
        )
        path_big = next(Path(path_run).glob("State_phys_*/state_phys*"))
        path_big_old = path_big.with_name("old_" + path_big.name)
        path_big.rename(path_big_old)

        # Then, the alternative implementation
        modif_resolution_from_dir_memory_efficient(
            path_run, t_approx, coef_modif_resol
        )

        with h5py.File(path_big_old, "r") as file:
            group_state_phys = file["/state_phys"]
            key = list(group_state_phys.keys())[0]
            field_old = group_state_phys[key][...]

        with h5py.File(path_big, "r") as file:
            group_state_phys = file["/state_phys"]
            field = group_state_phys[key][...]

        assert np.allclose(field_old, field)


class TestModifResol2d(TestModifResol3d):
    @classproperty
    def Simul(cls):
        from fluidsim.solvers.ns2d.solver import Simul

        return Simul



from fluidsim.base.output import OutputBasePseudoSpectral


class Output(OutputBasePseudoSpectral):
    """Output for spherical harmo solvers."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""

        OutputBasePseudoSpectral._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        classes.PhysFields.module_name = (
            "fluidsim.base.sphericalharmo.phys_fields"
        )
        classes.PhysFields.class_name = "PhysFieldsSphericalHarmo"



"""Operators finite differences (:mod:`fluidsim.operators.op_finitediff2d`)
===========================================================================

Provides:

.. autoclass:: OperatorFiniteDiff1DPeriodic
   :members:
   :private-members:

"""

import numpy as np

try:
    import scipy.sparse as sparse
except ImportError:
    pass

from .base import OperatorsBase1D


class OperatorFiniteDiff1DPeriodic(OperatorsBase1D):
    def __init__(self, params=None):
        super().__init__(params)
        nx = self.nx
        self.nx_seq = nx
        dx = self.deltax

        self.sparse_px = sparse.diags(
            diagonals=[-np.ones(nx - 1), np.ones(nx - 1), -1, 1],
            offsets=[-1, 1, nx - 1, -(nx - 1)],
        )
        self.sparse_px = self.sparse_px / (2 * dx)

        self.sparse_pxx = sparse.diags(
            diagonals=[np.ones(nx - 1), -2 * np.ones(nx), np.ones(nx - 1), 1, 1],
            offsets=[-1, 0, 1, nx - 1, -(nx - 1)],
        )

        self.sparse_pxx = self.sparse_pxx / dx ** 2

    def px(self, a):
        return self.sparse_px.dot(a.flat)

    def pxx(self, a):
        return self.sparse_pxx.dot(a.flat)

    def identity(self):
        return sparse.identity(self.size)

    def produce_long_str_describing_oper(self):
        return super().produce_long_str_describing_oper("Finite difference")



import unittest

import matplotlib.pyplot as plt
from numpy.testing import assert_array_almost_equal

import fluiddyn.util.mpi as mpi
from fluidsim.util.testing import (
    TestSimulConserveOutput,
    classproperty,
    skip_if_no_fluidfft,
)


@skip_if_no_fluidfft
class TestSimulSW1L(TestSimulConserveOutput):
    zero = 1e-5

    @classproperty
    def Simul(cls):
        from fluidsim.solvers.sw1l.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):
        super().init_params()
        params = cls.params
        params.output.HAS_TO_SAVE = True
        params.output.ONLINE_PLOT_OK = True
        params.output.spatial_means.HAS_TO_PLOT_SAVED = True
        params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True
        params.output.spectra.HAS_TO_PLOT_SAVED = True
        params.output.increments.HAS_TO_PLOT_SAVED = True
        params.output.pdf.HAS_TO_PLOT_SAVED = True

        params.output.periods_save.spatial_means = 0.25
        params.output.periods_save.phys_fields = 0.25
        params.output.periods_save.spect_energy_budg = 0.2
        params.output.periods_save.spectra = 0.2
        params.output.periods_save.increments = 0.25
        params.output.periods_save.pdf = 0.25
        params.output.periods_save.time_signals_fft = 0.1

        params.f = 1e-4
        params.forcing.enable = True
        params.forcing.type = "waves"
        params.forcing.nkmin_forcing = 2
        params.forcing.nkmax_forcing = 4
        params.init_fields.type = "wave"
        params.oper.nx = 16
        params.oper.ny = 16

    def test_phys_fields(self):
        self.plot("phys_fields")

    def test_print_stdout(self):
        self.plot("print_stdout")

    def test_spatial_means(self):
        self.plot("spatial_means")
        spatial_means = self.sim.output.spatial_means
        spatial_means.compute_time_means()
        spatial_means.time_first_saved()
        spatial_means.time_last_saved()

    def test_increments(self):
        self.plot("increments")
        self.sim.output.increments.plot_Kolmo()

    def test_spectra(self):
        """Test spectra loading and plotting.

        .. TODO::

           Errors

           * if (nx, ny) = (16, 8)
             >>> E_K = dset_spectrum1Dkx_EK[it] + dset_spectrum1Dky_EK[it]
             E ValueError: operands could not be broadcast together with
             shapes (9,) (5,)
           * if periods_save = 0.25.
             >>> delta_i_plot = int(np.round(delta_t / delta_t_save))
             E ValueError: cannot convert float NaN to integer

        """
        self.get_results("spectra")
        if mpi.nb_proc == 1:
            spectra = self.sim.output.spectra
            spectra.plot1d()
            spectra.plot2d()
            # spectra.plot_diss()
            spectra.compute_lin_spectra()
            plt.cla()

    def test_spect_energy_budg(self):
        """Test spect_energy_budg loading and plotting.

        .. TODO::

           Errors

           * if periods_save = 0.25.
             >>> delta_i_plot = int(np.round(delta_t / delta_t_save))
             E ValueError: cannot convert float NaN to integer

        """
        self.get_results("spect_energy_budg")
        self.plot("spect_energy_budg")

    def test_time_signals_fft(self):
        self.plot("time_signals_fft")
        self.sim.output.time_signals_fft.plot_spectra()

    def test_pdf(self):
        self.plot("pdf")
        self.get_results("pdf")

    def test_energy_vs_spatial_means(self):
        """Verify energy saved by spatial_means module is the same."""
        if mpi.nb_proc > 1:
            mpi.comm.barrier()
        dict_results_print_stdout = self.get_results("print_stdout")
        df_spatial_means = self.get_results("spatial_means")

        # ignore last row to be comparable to print_stdout
        imax = -1 if len(df_spatial_means) > 1 else None
        assert_array_almost_equal(
            dict_results_print_stdout["E"],
            df_spatial_means.E[:imax].values,
            decimal=4,
        )

    def get_tendencies(self):
        tendencies_fft = self.tendencies_fft
        Fx_fft = tendencies_fft.get_var("ux_fft")
        Fy_fft = tendencies_fft.get_var("uy_fft")
        Feta_fft = tendencies_fft.get_var("eta_fft")

        return Fx_fft, Fy_fft, Feta_fft

    def test_enstrophy_conservation(self):
        """Verify that the enstrophy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        """
        oper = self.sim.oper
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        Fq_fft, _, _ = oper.qapamfft_from_uxuyetafft(Fx_fft, Fy_fft, Feta_fft)
        q_fft = self.sim.state.get_var("q_fft")

        T_q = (Fq_fft.conj() * q_fft + Fq_fft * q_fft.conj()).real / 2.0
        sum_T = oper.sum_wavenumbers(T_q)
        self.assertAlmostZero(sum_T, tolerance_warning=False)

    def test_energy_conservation(self):
        """Verify that the energy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state_phys = self.sim.state.state_phys
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        oper = self.sim.oper
        Fx = oper.ifft2(Fx_fft)
        Fy = oper.ifft2(Fy_fft)
        Feta = oper.ifft2(Feta_fft)
        A = (
            Feta * (ux ** 2 + uy ** 2) / 2
            + (1 + eta) * (ux * Fx + uy * Fy)
            + self.sim.params.c2 * eta * Feta
        )

        A_fft = oper.fft2(A)
        if mpi.rank == 0:
            self.assertAlmostZero(A_fft[0, 0], tolerance_warning=False)

    def test_state_methods(self):
        state_spect = self.sim.state.state_spect.copy()
        self.sim.state.statespect_from_statephys()
        state_spect2 = self.sim.state.state_spect
        assert id(state_spect) != id(state_spect2)
        assert_array_almost_equal(state_spect, state_spect2)

    def test_state_init_from_uxuyfft(self):
        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")

        # The following method assumes that the velocity only has a rotational
        # component
        self.sim.state.init_from_uxuyfft(ux_fft, uy_fft)

        # Should have no divergence
        div = self.sim.state.compute("div")
        self.assertAlmostZero(div.max())

    def test_state_compute(self):
        for key in ("q", "h", "Floc"):
            var_computed = self.sim.state.compute(key)


if __name__ == "__main__":
    unittest.main()



"""Base classes for writing solvers
===================================


Provides:

.. autosummary::
   :toctree:

   solvers
   params
   setofvariables
   state
   init_fields
   time_stepping
   output
   forcing
   preprocess
   sphericalharmo

Third party solver interfaces:

.. autosummary::
   :toctree:

   basilisk
   dedalus

"""



"""
Plate2d forcing (:mod:`fluidsim.solvers.plate2d.forcing`)
=========================================================

"""

from fluidsim.base.forcing import ForcingBasePseudoSpectral

from fluidsim.base.forcing.specific import Proportional as ProportionalBase

from fluidsim.base.forcing.specific import (
    InScriptForcingPseudoSpectral,
    TimeCorrelatedRandomPseudoSpectral as TCRandomPS,
)


class ForcingPlate2D(ForcingBasePseudoSpectral):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        classes = [Proportional, TCRandomPSW, InScriptForcingPseudoSpectral]
        ForcingBasePseudoSpectral._complete_info_solver(info_solver, classes)


class SpecificForcingPlate2d:
    @classmethod
    def _modify_sim_repr_maker(cls, sim_repr_maker):
        sim = sim_repr_maker.sim
        sim_repr_maker.add_parameters(
            {"P": sim.params.forcing.forcing_rate},
            formats={"P": "5.0g"},
            indices={"P": 2},
        )


class TCRandomPSW(SpecificForcingPlate2d, TCRandomPS):
    _key_forced_default = "w_fft"


class Proportional(SpecificForcingPlate2d, ProportionalBase):
    _key_forced_default = "w_fft"


# class TimeCorrelatedRandomPseudoSpectral(TCRandomPS):
#     def compute_forcingc_raw(self):
#         Fw_fft = super(TimeCorrelatedRandomPseudoSpectral,
#                        self).compute_forcingc_raw()

#         return Fw_fft

# def compute_forcing_2nd_degree_eq(self):
#     """compute a forcing normalize with a 2nd degree eq."""

#     w_fft = self.sim.state.state_spect.get_var('w_fft')
#     vmax = self.sim.params.forcing.vmax
#     n0 = w_fft.shape[0]
#     n1 = w_fft.shape[1]
#     zf = np.zeros((n0, n1), dtype=np.complex128)
#     kfSI = 2*np.pi*2.5
#     sk = 1.
#     kfor = self.sim.oper.Lx*kfSI/(2*np.pi)
#     rand = np.random.normal(loc=0.0, scale=1.0, size=(2*n0/3, 2*n1/3))
#     for i0 in xrange(2*n0/3):
#         for i1 in xrange(2*n1/3):
#             zf[i0, i1] = ((1/(2*sk**2)) *
#                           (np.exp(-np.sqrt(i0**2+i1**2)-kfor-1)**2) *
#                           np.exp(2*np.pi*1j*rand[i0, i1]))

#             # zIJ(1:(n3+1),1:(n3+1)) =
#             # exp(-(sqrt(I.^2+J.^2)-kfor-1).^2/(2*sk^2))
#             # .*exp(2*pi*1i*rand(size(I)));
#     forcingW_fft = self.oper.nx_loc**2 * vmax / np.sum(zf) * zf
#     self.forcingc_fft.set_var('w_fft', forcingW_fft)

#     self.put_forcingc_in_forcing()



"""Preprocessing for pseudo-spectral solvers (:mod:`fluidsim.base.preprocess.pseudo_spect`)
===========================================================================================

Provides:

.. autoclass:: PreprocessPseudoSpectral
   :members:
   :private-members:


"""

import numpy as np
from .base import PreprocessBase


class PreprocessPseudoSpectral(PreprocessBase):
    _tag = "pseudo_spectral"

    def __call__(self):
        """Preprocesses if enabled."""

        super().__call__()
        if self.params.enable:
            if self.sim.params.forcing.enable:
                if "forcing" in self.params.init_field_scale:
                    self.set_forcing_rate()
                    self.normalize_init_fields()
                else:
                    self.normalize_init_fields()
                    self.set_forcing_rate()
            else:
                self.normalize_init_fields()

            self.sim.state.clear_computed()
            self.set_viscosity()
            self.output._save_info_solver_params_xml(replace=True)

    def normalize_init_fields(self):
        """
        A non-dimensionalization procedure for the initialized fields.

        Parameters
        ----------------
        init_field_scale : string (use 'energy', 'unity')
            Set quantity to normalize initialized fields with.
        """
        state = self.sim.state
        scale = self.params.init_field_scale
        C = float(self.params.init_field_const)

        if scale == "energy":
            try:
                (Ek,) = self.output.compute_quad_energies()
            except:
                Ek = self.output.compute_energy()

            ux_fft = state.get_var("ux_fft")
            uy_fft = state.get_var("uy_fft")

            if Ek != 0.0:
                ux_fft = (C / Ek) ** 0.5 * ux_fft
                uy_fft = (C / Ek) ** 0.5 * uy_fft

            try:
                state.init_from_uxuyfft(ux_fft, uy_fft)
            except AttributeError:
                rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
                state.init_statespect_from(rot_fft=rot_fft)
        elif scale == "enstrophy":
            omega_0 = self.output.compute_enstrophy()
            rot_fft = state.get_var("rot_fft")

            if omega_0 != 0.0:
                rot_fft = (C / omega_0) ** 0.5 * rot_fft
                state.init_from_rotfft(rot_fft)

        elif scale == "enstrophy_forcing":
            P = self.sim.params.forcing.forcing_rate
            k_f = self.oper.deltak * (
                (
                    self.sim.params.forcing.nkmax_forcing
                    + self.sim.params.forcing.nkmin_forcing
                )
                // 2
            )
            omega_0 = self.output.compute_enstrophy()
            rot_fft = state.get_var("rot_fft")

            if omega_0 != 0.0:
                C_0 = omega_0 / (P ** (2.0 / 3) * k_f ** (4.0 / 3))
                rot_fft = (C / C_0) ** 0.5 * rot_fft
                state.init_from_rotfft(rot_fft)

        elif scale == "unity":
            pass
        else:
            raise ValueError("Unknown initial fields scaling: ", scale)

    def set_viscosity(self):
        r"""Based on

        - the initial total enstrophy, \Omega_0, or

        - the initial energy

        - the forcing rate, \epsilon

        the viscosity scale or Reynolds number is set.

        Parameters
        ----------

        params.preprocess.viscosity_type : string
          Type/Order of viscosity desired

        params.preprocess.viscosity_scale : string
          Mean quantity to be scaled against

        params.preprocess.viscosity_const : float
          Calibration constant to set dissipative wave number

        Notes
        -----

        Algorithm: Sets viscosity variable nu and reinitializes f_d array for
        timestepping

        """
        params = self.params
        viscosity_type = params.viscosity_type
        viscosity_scale = params.viscosity_scale
        C = params.viscosity_const

        if viscosity_scale == "enstrophy":
            args = [self.output.compute_enstrophy()]
        elif viscosity_scale == "energy":
            args = [self.output.compute_energy()]
        elif viscosity_scale == "enstrophy_forcing":
            args = [
                self.output.compute_enstrophy(),
                self.sim.params.forcing.forcing_rate,
            ]
        elif viscosity_scale == "forcing":
            args = [self.sim.params.forcing.forcing_rate]
        else:
            raise ValueError("Unknown viscosity scale: %s" % viscosity_scale)

        result = calcul_viscosity(
            C,
            viscosity_scale,
            viscosity_type,
            oper=self.oper,
            verbose=False,
            *args,
        )
        for v in result.values():
            attr, order, nu = v
            self.sim.params.__setattr__(attr, nu)

        self.sim.time_stepping.__init__(self.sim)

    def set_forcing_rate(self):
        r"""Based on C, a non-dimensional ratio of forcing rate to one of the
        following forcing scales

        - the initial total energy, math:: E_0
        - the initial total enstrophy, math:: \Omega_0

        the forcing rate is set.

        Parameters
        ----------

        params.preprocess.forcing_const : float
          Non-dimensional ratio of forcing_scale to forcing_rate

        params.preprocess.forcing_scale : string
          Mean quantity to be scaled against

        """
        params = self.params

        forcing_scale = params.forcing_scale
        C = float(params.forcing_const)
        # Forcing wavenumber
        k_f = self.oper.deltak * (
            (
                self.sim.params.forcing.nkmax_forcing
                + self.sim.params.forcing.nkmin_forcing
            )
            // 2
        )

        if forcing_scale == "unity":
            self.sim.params.forcing.forcing_rate = C
        elif forcing_scale == "energy":
            energy_0 = self.output.compute_energy()
            self.sim.params.forcing.forcing_rate = C * energy_0 ** 1.5 * k_f
        elif forcing_scale == "enstrophy":
            omega_0 = self.output.compute_enstrophy()
            self.sim.params.forcing.forcing_rate = C * omega_0 ** 1.5 / k_f ** 2
        else:
            raise ValueError("Unknown forcing scale: %s" % forcing_scale)

        self.sim.forcing.__init__(self.sim)


def calcul_viscosity(
    viscosity_const, viscosity_scale, viscosity_type, *args, **kwargs
):
    """Calculates viscosity based on scaling formulae.  Use this function
    to estimate viscosity before runtime.

    Parameters
    ----------
    viscosity_const : scalar
      Calibration constant to set dissipative wave number

    viscosity_scale : string
      Mean quantity to be scaled against

    viscosity_type : string
      Type/Order of viscosity desired

    *args : scalar(s)
      Estimated value for viscosity scale

    oper : object of :class:`fluidsim.operators.operators.Operators`, optional

    nh, Lh, coef_dealiasing, nk_f : scalars, optional
        No. of grid points, length of the box, coeff. of dealiasing and
        forcing wavenumber index.

    verbose : bool, optional
        For verbose output


    Examples
    --------
    >>> calcul_viscosity(
    ...     1, 'enstrophy', 'laplacian', 50., oper=sim.oper)
    >>> calcul_viscosity(
    ...     0.5, 'forcing', 'laplacian_hyper8', oper=sim.oper, verbose=False)
    >>> calcul_viscosity(
    ...     1, 'enstrophy_forcing', 'laplacian', 50., 1.,
    ...     nh=128, Lh=50, coef_dealiasing=2. / 3, nk_f=6)
    >>> calcul_viscosity(
    ...     0.785, 'forcing', 'hyper8', 1.,
    ...     nh=1920, Lh=50, coef_dealiasing=8. / 9, nk_f=6)


    """
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    else:
        verbose = True

    if "oper" in kwargs:
        oper = kwargs["oper"]
        coef_dealiasing = oper.coef_dealiasing
        nk_f = (
            oper.params.forcing.nkmax_forcing + oper.params.forcing.nkmin_forcing
        ) // 2
        delta_x = oper.deltax
        deltak = oper.deltak
    else:
        nh = kwargs["nh"]
        Lh = kwargs["Lh"]
        coef_dealiasing = kwargs["coef_dealiasing"]
        nk_f = kwargs["nk_f"]
        delta_x = Lh / nh
        deltak = 2 * np.pi / Lh

    # Smallest resolved scale
    k_max = np.pi / delta_x * coef_dealiasing
    # OR np.pi / k_d, the dissipative wave number
    C = viscosity_const
    length_scale = C * np.pi / k_max
    k_f = deltak * nk_f
    large_scale = np.pi / k_f

    k_diss = k_max / C / np.pi
    if verbose:
        print("Max. wavenumber =", np.pi / delta_x)
        print("Max. resolved wavenumber, k_max =", k_max)
        print("Grid spacing, delta_x =", delta_x)
        print("\nESTIMATED (P~eps)")
        print(
            f"Dissipation wavenumber, k_d = {k_diss}; k_d / k_f = {k_diss / k_f}"
        )
        print(
            f"Dissipation length scale, L_d = {length_scale}; L_d / L_f = {length_scale / large_scale}"
        )
        print("Viscosity scale:", viscosity_scale, "=", args)

    if len(args) == 0:
        raise ValueError("Expected values related to `viscosity_scale` as *args")

    if viscosity_scale == "enstrophy":
        omega_0 = args[0]
        eta = omega_0 ** 1.5
        time_scale = eta ** (-1.0 / 3)
    elif viscosity_scale == "energy":
        energy_0 = args[0]
        epsilon = energy_0 * (1.5) / large_scale
        time_scale = epsilon ** (-1.0 / 3) * length_scale ** (2.0 / 3)
    elif viscosity_scale == "enstrophy_forcing":
        omega_0 = args[0]
        eta = omega_0 ** 1.5
        t1 = eta ** (-1.0 / 3)
        # Energy dissipation rate
        epsilon = args[1]
        t2 = epsilon ** (-1.0 / 3) * length_scale ** (2.0 / 3)
        time_scale = min(t1, t2)
    elif viscosity_scale == "forcing":
        epsilon = args[0]
        time_scale = epsilon ** (-1.0 / 3) * length_scale ** (2.0 / 3)
    else:
        raise ValueError("Unknown viscosity scale: %s" % viscosity_scale)

    dict_visc = {
        "laplacian": ["nu_2", 2],
        "hyper4": ["nu_4", 4],
        "hyper8": ["nu_8", 8],
        "hypo": ["nu_m4", -4],
    }

    if verbose:
        if "oper" in kwargs:
            epsilon = oper.params.forcing.forcing_rate
        elif "eps" in kwargs:
            epsilon = kwargs["eps"]
        else:
            epsilon = 1.0

        print("Dissipation, epsilon =", epsilon)
        kolmo_len = []

    if not any([k in viscosity_type for k in dict_visc]):
        raise ValueError("Unknown viscosity type: %s" % viscosity_type)

    else:
        for k, v in dict_visc.items():
            if k in viscosity_type:
                attr, order = v
                nu = length_scale ** order / time_scale
                v.append(nu)
                if verbose:
                    kolmo_len.append(
                        (nu ** 3 / epsilon) ** (1.0 / (3 * order - 2))
                    )
            else:
                v.append(0.0)

    if verbose:
        length_scale = np.mean(kolmo_len)
        k_diss = 1.0 / length_scale
        print(f"\nCALCULATED (eps={epsilon})")
        print(
            f"Dissipation wavenumber, k_d = {k_diss}; k_d / k_f = {k_diss / k_f}"
        )
        print(
            f"Dissipation length scale, L_d = {length_scale}; L_d / L_f = {length_scale / large_scale}"
        )

    return dict_visc



#!/usr/bin/env python
# mpirun -np 4 python simul_force_vortex_grid.py

from math import pi
from fluidsim.solvers.sw1l.solver import Simul
from fluidsim.util.util import mpi

params = Simul.create_default_params()
params.short_name_type_run = "vortex_grid"
params.output.sub_directory = "beskow_tests"

# --------Grid parameters---------
params.oper.nx = params.oper.ny = nh = 512
params.oper.Lx = params.oper.Ly = Lh = 2 * pi
delta_x = Lh / nh

# -----Numerical method parameters----
params.f = 0.0
params.c2 = 100
params.oper.coef_dealiasing = 2.0 / 3
params.time_stepping.t_end = 100.0
params.init_fields.type = "vortex_grid"

# ----- Forcing parameters -----------
params.FORCING = True
params.forcing.type = "waves"
params.forcing.nkmax_forcing = 8
params.forcing.nkmin_forcing = 5
params.forcing.forcing_rate = 1.0

# k_max = pi / delta_x * params.oper.coef_dealiasing # Smallest resolved scale
# k_d = k_max / 2.5
# length_scale = pi / k_d
# params.nu_8 = params.forcing.forcing_rate**(1./3) * length_scale ** (22./3)

# ------Save file / StdOut parameters------------
params.output.periods_print.print_stdout = 0.1
params.output.period_refresh_plots = 0.5
params.output.periods_plot.phys_fields = 0.0
# params.output.spatial_means.HAS_TO_PLOT_SAVED = True
# params.output.spectra.HAS_TO_PLOT_SAVED = True
# params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True

params.output.periods_save.phys_fields = 1.0
params.output.periods_save.spectra = 0.1
params.output.periods_save.spatial_means = 0.1
params.output.periods_save.spect_energy_budg = 0.5
params.output.periods_save.increments = 0.5

# -----Preprocess parameters------------
params.preprocess.enable = True
params.preprocess.forcing_const = 5e4
# params.preprocess.forcing_scale = 'unity'
params.preprocess.forcing_scale = "enstrophy"
params.preprocess.viscosity_const = 1.5
params.preprocess.viscosity_scale = "enstrophy_forcing"
params.preprocess.viscosity_type = "hyper8"

sim = Simul(params)
sim.output.print_stdout(
    "Froude number ~ {:3e}, Viscosity = {:3e} + {:3e}".format(
        (1.0 / params.c2), sim.params.nu_2, sim.params.nu_8
    )
)
sim.output.print_stdout(f"Forcing rate = {sim.params.forcing.forcing_rate}")

sim.time_stepping.start()



""" """

import numpy as np

from fluidsim.solvers.sw1l.output import OutputSW1L


class OutputSW1LWaves(OutputSW1L):
    """subclass :class:`OutputSW1L`"""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        OutputSW1L._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        classes.SpectralEnergyBudget.class_name = (
            "SpectralEnergyBudgetSW1L"  # Waves'
        )



"""Modified SW1L equations solving exactly the linear terms
===========================================================

(:mod:`fluidsim.solvers.sw1l.exactlin.modified.solver`)



"""

import numpy as np

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.solvers.sw1l.solver import InfoSolverSW1L
from fluidsim.solvers.sw1l.exactlin.solver import Simul as SimulSW1LExactLin


class InfoSolverSW1LExactLinModified(InfoSolverSW1L):
    """Information about the solver SW1L."""

    def _init_root(self):
        super()._init_root()

        sw1l = "fluidsim.solvers.sw1l"

        self.module_name = sw1l + ".exactlin.modified.solver"
        self.class_name = "Simul"
        self.short_name = "SW1Lexmod"

        classes = self.classes

        classes.State.module_name = sw1l + ".exactlin.state"
        classes.State.class_name = "StateSW1LExactLin"

        classes.Output.module_name = sw1l + ".exactlin.modified.output"
        classes.Output.class_name = "OutputSW1LExactlinModified"


class Simul(SimulSW1LExactLin):
    """A solver of the shallow-water 1 layer equations (SW1L)"""

    InfoSolver = InfoSolverSW1LExactLinModified

    def tendencies_nonlin(self, state_spect=None, old=None):
        oper = self.oper
        fft2 = oper.fft2
        ifft2 = oper.ifft2

        if state_spect is None:
            state_phys = self.state.state_phys
            state_spect = self.state.state_spect
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        # compute the nonlinear terms for ux, uy and eta
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        rot_fft = self.state.get_var("rot_fft")
        ux_rot_fft, uy_rot_fft = oper.vecfft_from_rotfft(rot_fft)
        ux_rot = ifft2(ux_rot_fft)
        uy_rot = ifft2(uy_rot_fft)

        N1x_fft = fft2(ux_rot * ux)
        N2x_fft = fft2(uy_rot * ux)
        N1y_fft = fft2(ux_rot * uy)
        N2y_fft = fft2(uy_rot * uy)

        Nx_fft = -oper.divfft_from_vecfft(N1x_fft, N2x_fft)
        Ny_fft = -oper.divfft_from_vecfft(N1y_fft, N2y_fft)

        jx_fft = fft2(ux_rot * eta)
        jy_fft = fft2(uy_rot * eta)
        Neta_fft = -oper.divfft_from_vecfft(jx_fft, jy_fft)

        # self.verify_tendencies(state_spect, state_phys, Nx_fft, Ny_fft, Neta_fft)

        # compute the nonlinear terms for q, ap and am
        (Nq_fft, Np_fft, Nm_fft) = oper.qapamfft_from_uxuyetafft(
            Nx_fft, Ny_fft, Neta_fft
        )

        oper.dealiasing(Nq_fft, Np_fft, Nm_fft)

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old
        tendencies_fft.set_var("q_fft", Nq_fft)
        tendencies_fft.set_var("ap_fft", Np_fft)
        tendencies_fft.set_var("am_fft", Nm_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft

    # def verify_tendencies(
    #     self, state_spect, state_phys, Nx_fft, Ny_fft, Neta_fft
    # ):
    #     """For verifying conservation of quadratic energy."""

    #     oper = self.oper
    #     ux_fft = self.state.get_var("ux_fft")
    #     uy_fft = self.state.get_var("uy_fft")
    #     eta_fft = self.state.get_var("eta_fft")

    #     oper.dealiasing(Nx_fft, Ny_fft, Neta_fft)
    #     T_ux = (ux_fft.conj() * Nx_fft).real
    #     T_uy = (uy_fft.conj() * Ny_fft).real
    #     T_eta = (eta_fft.conj() * Neta_fft).real * self.params.c2
    #     T_tot = T_ux + T_uy + T_eta
    #     print(
    #         "sum(T_tot) = {0:9.4e} ; sum(abs(T_tot)) = {1:9.4e}".format(
    #             self.oper.sum_wavenumbers(T_tot),
    #             self.oper.sum_wavenumbers(abs(T_tot)),
    #         )
    #     )


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 64
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 2.0

    params.init_fields.type = "vortex_grid"

    params.forcing.enable = True
    params.forcing.type = "waves"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 1.0
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spect_energy_budg = 0.5
    params.output.periods_save.increments = 0.5
    params.output.periods_save.pdf = 0.5
    params.output.periods_save.time_signals_fft = False

    params.output.periods_plot.phys_fields = 0.0

    params.output.phys_fields.field_to_plot = "div"

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    # sim.output.phys_fields.plot()

    fld.show()



# -*- coding: utf-8 -*-
"""Output (:mod:`fluidsim.solvers.ns2d.strat.output`)
=====================================================

Provides the modules:

.. autosummary::
   :toctree:

   print_stdout
   spatial_means
   spectra
   spect_energy_budget

and the main output class for the ns2d.strat solver:

.. autoclass:: OutputStrat
   :members:
   :private-members:

"""

import numpy as np

from math import radians

from fluidsim.solvers.ns2d.output import Output


class OutputStrat(Output):
    """Output for ns2d.strat solver."""

    def __init__(self, sim):
        self.sim = sim
        if sim.params.forcing.type.endswith("anisotropic"):
            self._init_froude_number()
            self.ratio_omegas = self._compute_ratio_omegas()

        super().__init__(sim)

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""

        Output._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        base_name_mod = "fluidsim.solvers.ns2d.strat.output"

        classes.PrintStdOut.module_name = base_name_mod + ".print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOutNS2DStrat"

        classes.PhysFields.module_name = base_name_mod + ".phys_fields"
        classes.PhysFields.class_name = "PhysFields2DStrat"

        classes.Spectra._set_attribs(
            {
                "module_name": base_name_mod + ".spectra",
                "class_name": "SpectraNS2DStrat",
            }
        )

        classes.SpectraMultiDim._set_attribs(
            {
                "module_name": base_name_mod + ".spectra_multidim",
                "class_name": "SpectraMultiDimNS2DStrat",
            }
        )

        classes.SpatialMeans._set_attribs(
            {
                "module_name": base_name_mod + ".spatial_means",
                "class_name": "SpatialMeansNS2DStrat",
            }
        )

        classes.SpectEnergyBudg._set_attribs(
            {
                "module_name": base_name_mod + ".spect_energy_budget",
                "class_name": "SpectralEnergyBudgetNS2DStrat",
            }
        )

    def compute_energies_fft(self):
        """Compute the kinetic and potential energy (k)"""
        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        b_fft = self.sim.state.state_spect.get_var("b_fft")
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
        energyK_fft = (np.abs(ux_fft) ** 2 + np.abs(uy_fft) ** 2) / 2

        if self.sim.params.N == 0:
            energyA_fft = np.zeros_like(energyK_fft)
        else:
            energyA_fft = ((np.abs(b_fft) / self.sim.params.N) ** 2) / 2
        return energyK_fft, energyA_fft

    def compute_energies2_fft(self):
        """Compute the two kinetic energies for u_x and u_y"""
        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
        energyK_ux_fft = (np.abs(ux_fft) ** 2) / 2
        energyK_uy_fft = (np.abs(uy_fft) ** 2) / 2
        return energyK_ux_fft, energyK_uy_fft

    def compute_energy_fft(self):
        """Compute energy(k)"""
        energyK_fft, energyA_fft = self.compute_energies_fft()
        return energyK_fft + energyA_fft

    def compute_enstrophy_fft(self):
        """Compute enstrophy(k)"""
        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        return np.abs(rot_fft) ** 2 / 2

    def compute_energy(self):
        """Compute the spatially averaged energy."""
        energy_fft = self.compute_energy_fft()
        return self.sum_wavenumbers(energy_fft)

    def compute_energies(self):
        """Compute the kinetic and potential energy"""
        energyK_fft, energyA_fft = self.compute_energies_fft()
        energyK_ux_fft, energyK_uy_fft = self.compute_energies2_fft()
        energyK = self.sum_wavenumbers(energyK_fft)
        energyA = self.sum_wavenumbers(energyA_fft)

        # energyK_ux (##TODO: remove it. not sense)
        energyK_ux = self.sum_wavenumbers(energyK_ux_fft)
        return energyK, energyA, energyK_ux

    def compute_enstrophy(self):
        """Compute the spatially averaged enstrophy."""
        enstrophy_fft = self.compute_enstrophy_fft()
        return self.sum_wavenumbers(enstrophy_fft)

    def _init_froude_number(self):
        """Compute froude number ONLY for anisotropic forcing."""
        angle = self.sim.params.forcing.tcrandom_anisotropic.angle
        if isinstance(angle, str):
            if angle.endswith("°"):
                angle = radians(float(angle[:-1]))
            else:
                raise ValueError(
                    "Angle should be a string with \n"
                    + "the degree symbol or a float in radians"
                )

        self.froude_number = np.sin(angle)

    def _compute_ratio_omegas(self):
        r"""Compute ratio omegas; gamma = \omega_l / \omega_{af}"""
        params = self.sim.params
        pforcing = params.forcing
        forcing_rate = pforcing.forcing_rate

        # Compute forcing wave-number
        nkmax_forcing = pforcing.nkmax_forcing
        nkmin_forcing = pforcing.nkmin_forcing

        deltak = max(2 * np.pi / params.oper.Lx, 2 * np.pi / params.oper.Ly)
        k_f = ((nkmax_forcing + nkmin_forcing) / 2) * deltak
        l_f = 2 * np.pi / k_f

        # Compute linear frequency in s^-1
        omega_l = params.N * self.froude_number
        omega_l = omega_l / (2 * np.pi)

        # Compute forcing frequency
        if pforcing.normalized.constant_rate_of is None:
            if pforcing.key_forced is None or pforcing.key_forced == "rot_fft":
                omega_af = forcing_rate ** (1.0 / 3)

            elif pforcing.key_forced == "ap_fft":
                omega_af = forcing_rate ** (1.0 / 7) * l_f ** (-2.0 / 7)

            else:
                raise ValueError("params.forcing.key_forced is not known.")

        elif pforcing.normalized.constant_rate_of in ["energy", "energyK"]:
            omega_af = forcing_rate ** (1.0 / 3) * l_f ** (-2.0 / 3)

        else:
            raise ValueError(f"{pforcing.normalized.constant_rate_of} not known.")

        return omega_l / omega_af

    def _init_sim_repr_maker(self):
        """initialize sim_repr_maker."""
        sim_repr_maker = super()._init_sim_repr_maker()
        if self.sim.params.forcing.type.endswith("anisotropic"):
            sim_repr_maker.add_parameters(
                {"F": self.froude_number, "gamma": self.ratio_omegas},
                formats={"gamma": ".1f"},
            )
        return sim_repr_maker

    def plot_summary(self, field="b", time_phys=None, tmin=None, tmax=None):
        """
        Plots summary of all outputs of a simulation.
        """
        # pylint: disable=maybe-no-member
        self.phys_fields.plot(field=field, time=time_phys, QUIVER=False)
        self.print_stdout.plot()
        self.spatial_means.plot()
        self.spectra.plot1d(tmin=tmin, tmax=tmax, level3=10.0)
        self.spect_energy_budg.plot(tmin=tmin, tmax=tmax)



"""SW1L equations solving exactly the linear terms
==================================================

(:mod:`fluidsim.solvers.sw1l.onlywaves.solver`)

This class is a solver of the 1 layer shallow water (Saint Venant)
equations with zeros QG PV.
"""

import numpy as np

from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.solvers.sw1l.exactlin.solver import InfoSolverSW1LExactLin
from fluidsim.solvers.sw1l.exactlin.solver import Simul as SimulSW1LExactLin


from fluiddyn.util import mpi


class InfoSolverSW1LWaves(InfoSolverSW1LExactLin):
    """Information about the solver SW1L."""

    def _init_root(self):
        super()._init_root()

        sw1l = "fluidsim.solvers.sw1l"

        self.module_name = sw1l + ".onlywaves.solver"
        self.short_name = "SW1Lwaves"

        classes = self.classes

        classes.State.module_name = sw1l + ".onlywaves.state"
        classes.State.class_name = "StateSW1LWaves"

        classes.Output.module_name = sw1l + ".onlywaves.output"
        classes.Output.class_name = "OutputSW1LWaves"


class Simul(SimulSW1LExactLin):
    """A solver of the shallow-water 1 layer equations (SW1L)"""

    InfoSolver = InfoSolverSW1LWaves

    def tendencies_nonlin(self, state_spect=None, old=None):
        oper = self.oper
        fft2 = oper.fft2

        if state_spect is None:
            state_phys = self.state.state_phys
            state_spect = self.state.state_spect
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        # compute the nonlinear terms for ux, uy and eta
        gradu2_x_fft, gradu2_y_fft = oper.gradfft_from_fft(
            fft2(ux ** 2 + uy ** 2) / 2
        )

        Nx_fft = -gradu2_x_fft
        Ny_fft = -gradu2_y_fft

        if self.params.f > 0:
            # not very efficient, but simple...
            rot = self.state.get_var("rot")
            N1x = +rot * uy
            N1y = -rot * ux

            Nx_fft += fft2(N1x)
            Ny_fft += fft2(N1y)

        jx_fft = fft2(eta * ux)
        jy_fft = fft2(eta * uy)
        Neta_fft = -oper.divfft_from_vecfft(jx_fft, jy_fft)

        # self.verify_tendencies(state_spect, state_phys,
        #                        Nx_fft, Ny_fft, Neta_fft)

        # compute the nonlinear terms for q, ap and am
        (Nq_fft, Np_fft, Nm_fft) = self.oper.qapamfft_from_uxuyetafft(
            Nx_fft, Ny_fft, Neta_fft
        )

        # Np_fft = self.oper.create_arrayK(value=0)
        # Nm_fft = self.oper.create_arrayK(value=0)

        oper.dealiasing(Np_fft, Nm_fft)

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old
        tendencies_fft.set_var("ap_fft", Np_fft)
        tendencies_fft.set_var("am_fft", Nm_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft

    def compute_freq_complex(self, key):
        K2 = self.oper.K2
        # return self.oper.create_arrayK(value=0)
        if key == "ap_fft":
            omega = 1.0j * np.sqrt(self.params.f ** 2 + self.params.c2 * K2)
        elif key == "am_fft":
            omega = -1.0j * np.sqrt(self.params.f ** 2 + self.params.c2 * K2)
        return omega

    # def verify_tendencies(
    #     self, state_spect, state_phys, Nx_fft, Ny_fft, Neta_fft
    # ):
    #     # for verification conservation energy
    #     # compute the linear terms
    #     oper = self.oper
    #     ux = state_phys.get_var("ux")
    #     uy = state_phys.get_var("uy")
    #     eta = state_phys.get_var("eta")

    #     # q_fft = self.oper.create_arrayK(value=0)
    #     ap_fft = state_spect.get_var("ap_fft")
    #     am_fft = state_spect.get_var("am_fft")
    #     a_fft = ap_fft + am_fft
    #     div_fft = self.divfft_from_apamfft(ap_fft, am_fft)

    #     eta_fft = oper.etafft_from_afft(a_fft)

    #     dx_c2eta_fft, dy_c2eta_fft = oper.gradfft_from_fft(
    #         self.params.c2 * eta_fft
    #     )
    #     LCx = self.params.f * uy
    #     LCy = -self.params.f * ux
    #     Lx_fft = oper.fft2(LCx) - dx_c2eta_fft
    #     Ly_fft = oper.fft2(LCy) - dy_c2eta_fft
    #     Leta_fft = -div_fft

    #     # compute the full tendencies
    #     Fx_fft = Lx_fft + Nx_fft
    #     Fy_fft = Ly_fft + Ny_fft
    #     Feta_fft = Leta_fft + Neta_fft
    #     oper.dealiasing(Fx_fft, Fy_fft, Feta_fft)

    #     # test : ux, uy, eta ---> q, ap, am
    #     (Fq_fft, Fp_fft, Fm_fft) = self.oper.qapamfft_from_uxuyetafft(
    #         Fx_fft, Fy_fft, Feta_fft
    #     )
    #     # test : q, ap, am ---> ux, uy, eta
    #     (Fx2_fft, Fy2_fft, Feta2_fft) = self.oper.uxuyetafft_from_qapamfft(
    #         Fq_fft, Fp_fft, Fm_fft
    #     )
    #     print(np.max(abs(Fx2_fft - Fx_fft)))
    #     print(np.max(abs(Fy2_fft - Fy_fft)))
    #     print(np.max(abs(Feta2_fft - Feta_fft)))
    #     Fx_fft = Fx2_fft
    #     Fy_fft = Fy2_fft
    #     Feta_fft = Feta2_fft

    #     (Fq2_fft, Fp2_fft, Fm2_fft) = self.oper.qapamfft_from_uxuyetafft(
    #         Fx2_fft, Fy2_fft, Feta2_fft
    #     )
    #     print(np.max(abs(Fq2_fft - Fq_fft)))
    #     print(np.max(abs(Fp2_fft - Fp_fft)))
    #     print(np.max(abs(Fm2_fft - Fm_fft)))

    #     Fx = oper.ifft2(Fx_fft)
    #     Fy = oper.ifft2(Fy_fft)
    #     Feta = oper.ifft2(Feta_fft)
    #     A = (
    #         Feta * (ux ** 2 + uy ** 2) / 2
    #         + (1 + eta) * (ux * Fx + uy * Fy)
    #         + self.params.c2 * eta * Feta
    #     )
    #     A_fft = oper.fft2(A)
    #     if mpi.rank == 0:
    #         print("should be zero =", A_fft[0, 0])


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 64
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 2.0

    params.init_fields.type = "noise"

    params.forcing.enable = True
    params.forcing.type = "waves"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 1.0
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spect_energy_budg = 0.5
    params.output.periods_save.increments = 0.5
    params.output.periods_save.pdf = 0.5
    params.output.periods_save.time_signals_fft = False

    params.output.periods_plot.phys_fields = 0.0

    params.output.phys_fields.field_to_plot = "div"

    sim = Simul(params)

    sim.output.phys_fields.plot()
    sim.time_stepping.start()
    sim.output.phys_fields.plot()

    fld.show()



"""Run profiles (:mod:`fluidsim.util.console.profile`)
======================================================

"""

import gc
import os
from time import time
from pathlib import Path

import pstats
import cProfile

import numpy as np
import matplotlib.pyplot as plt

from fluiddyn.util import mpi
from fluiddyn.io import stdout_redirected

from fluidsim import _is_testing

from ..util import import_module_solver_from_key
from .util import (
    modif_params2d,
    modif_params3d,
    init_parser_base,
    parse_args_dim,
    tear_down,
    get_path_file,
)

from .bench import get_opfft


path_results = "/tmp/fluidsim_profile"
old_print = print
print = mpi.printby0
rank = mpi.rank
nb_proc = mpi.nb_proc
description = "Profile time-elapsed in various function calls"


def run_profile(sim, nb_dim=None, path_results=None, plot=False, verbose=True):
    """Profile a simulation run and save the results in `profile.pstats`

    Parameters
    ----------
    sim : Simul
        An initialized simulation object
    nb_dim : int
        Dimension of the solver
    path_results : str
        Path where all pstats files will be saved

    """
    if nb_dim is None:
        try:
            nb_dim = len(sim.oper.axes)
        except AttributeError:
            raise ValueError

    if path_results is None:
        path_results = Path.cwd()
    else:
        path_results = Path(path_results)

    if path_results.name.endswith(".pstats"):
        path = path_results
        path.parent.mkdir(parents=True, exist_ok=True)
    else:
        path, t_as_str = get_path_file(sim, path_results, "profile", ".pstats")
        path = Path(path)

    if mpi.rank > 0:
        path = path.with_name(path.stem + f"_rank{mpi.rank}" + path.suffix)

    with stdout_redirected(not verbose):
        sim.time_stepping.init_from_params()
        t0 = time()
        cProfile.runctx(
            "sim.time_stepping.main_loop()", globals(), locals(), path
        )
        t_end = time()
        sim.time_stepping.finalize_main_loop()

    if mpi.rank == 0:
        times = analyze_stats(path, nb_dim, plot)
        print(f"\nelapsed time = {t_end - t0:.3f} s")
        name_solver = sim.__module__.rsplit(".", maxsplit=1)[0]
        print(
            f"To display the results:\nfluidsim-profile -p -sf {path} -s {name_solver}"
        )
        print(
            "\nwith gprof2dot and graphviz (command dot):\n"
            f"gprof2dot -f pstats {path} | dot -Tpng -o profile.png"
        )

    else:
        # Retain only rank 0 profiles
        os.remove(path)

    return path


def profile(
    solver,
    dim="2d",
    n0=1024 * 2,
    n1=None,
    n2=None,
    path_dir=None,
    type_fft=None,
    verbose=False,
    plot=False,
    it_end=None,
):
    """Instantiate simulation object and run profiles."""

    inputs = dict(solver=solver, n0=n0, n1=n1, type_fft=type_fft)

    if dim == "3d":
        inputs["n2"] = n2

    print("profile with inputs:")
    for k, v in inputs.items():
        print(k, "=", str(v))

    def _profile(type_fft):
        """Run profile once."""
        Simul = solver.Simul
        params = Simul.create_default_params()

        if dim == "2d":
            modif_params2d(
                params, n0, n1, name_run="bench", type_fft=type_fft, it_end=it_end
            )
        elif dim == "3d":
            modif_params3d(
                params,
                n0,
                n1,
                n2,
                name_run="bench",
                type_fft=type_fft,
                it_end=it_end,
            )
        else:
            raise ValueError("dim has to be in ['2d', '3d']")

        nb_dim = int(dim[0])

        with stdout_redirected(not verbose):
            sim = Simul(params)

        try:
            run_profile(sim, nb_dim, path_dir, verbose=verbose)
        except Exception as e:
            if _is_testing:
                raise

            else:
                print(
                    "WARNING: Some error occurred while running benchmark"
                    " / saving results!"
                )
                raise

                print(e)
        finally:
            tear_down(sim)
            gc.collect()

    if str(type_fft).lower() == "all":
        d = get_opfft(n0, n1, n2, dim, only_dict=True)
        for type_fft, cls in d.items():
            if cls is not None:
                print(type_fft)
                _profile(type_fft)
    else:
        _profile(type_fft)


def init_parser(parser):
    """Initialize argument parser for `fluidsim profile`."""

    init_parser_base(parser)
    parser.add_argument("-o", "--output_dir", default=path_results)
    parser.add_argument(
        "-t",
        "--type-fft",
        default=None,
        help=(
            'specify FFT type key (for eg. "fft2d.mpi_with_fftw1d") or "all";'
            "if not specified uses the default FFT method in operators"
        ),
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-sf", "--stats_file", default=None)
    parser.add_argument("-p", "--plot", action="store_true")
    parser.add_argument(
        "-it", "--it-end", default=10, type=int, help="Number of iterations"
    )


def run(args):
    """Run `fluidsim profile` command."""

    if args.stats_file is not None and args.solver is None:
        # get solver from name... Not clean at all...
        tmp = os.path.split(args.stats_file)[-1]
        args.solver = tmp.split("result_profile_")[-1].split("_")[0]

    args = parse_args_dim(args)

    if args.stats_file is not None:
        nb_dim = int(args.dim.split("d")[0])
        analyze_stats(args.stats_file, nb_dim, args.plot)
        return

    # Initialize simulation and run benchmarks
    solver = import_module_solver_from_key(args.solver)
    profile(
        solver,
        args.dim,
        args.n0,
        args.n1,
        args.n2,
        path_dir=args.output_dir,
        type_fft=args.type_fft,
        verbose=args.verbose,
        plot=args.plot,
        it_end=args.it_end,
    )


def _compute_shorter_name(key, kind):
    if kind == "fft_as":
        words = key.split()
        if len(words) == 1:
            return key

        else:
            return words[1][1:-1]

    if kind == "pythran" in key:
        if "_pythran." in key:
            key = key.split("_pythran.")[-1].rstrip(">") + " (pythran)"
        else:
            key = key.split("<built-in method ")[-1].rstrip(">")
        return key

    if "fluidsim.solvers." in key:
        return key.split("fluidsim.solvers.")[-1][:-1]

    if key.startswith("<built-in method "):
        return key.split("<built-in method ")[-1][:-1]

    if key.startswith("<method "):
        key = key.split("<method ")[-1][:-1]
        return key.split(" ")[0].strip("'")

    return key


def plot_pie(
    times,
    long_functions,
    ax=None,
    times_descending=False,
    for_latex=False,
    **kwargs,
):
    """Plot a pie chart """
    percentages = []
    labels = []
    for k, v in long_functions.items():
        percentages.append(v["percentage"])
        name = _compute_shorter_name(k, v["kind"])
        if for_latex:
            name = r"\texttt{" + name.replace("_", r"\_") + r"}"
        labels.append(name)

    percentages.append(100 - sum(percentages))
    percentages = np.array(percentages)
    labels.append("other")

    labels = np.array(labels)
    if times_descending:
        args = percentages.argsort()[::-1]
        percentages.sort()
        percentages = percentages[::-1]
    else:
        args = percentages.argsort()
        percentages.sort()

    labels = labels[args]
    if "explode" in kwargs:
        if kwargs["explode"] == "other":
            # Explode the 'other' slice from the pie chart
            idx_other = np.argwhere(labels == "other")[0]
            explode = np.zeros_like(labels, dtype=float)
            explode[idx_other] = 0.2
            kwargs["explode"] = explode
        elif isinstance(kwargs["explode"], float):
            # Explode all slices based on percentage
            kwargs["explode"] /= percentages

    if "labeldistance" in kwargs:
        if isinstance(kwargs["labeldistance"], float):
            kwargs["labeldistance"] /= percentages
            kwargs["labeldistance"] += 1.1

    if ax is None:
        fig, ax = plt.subplots()

    if "startangle" not in kwargs:
        kwargs["startangle"] = 0

    for label, perc in zip(labels, percentages):
        print(
            "(label, perc) = ({:40s} {:5.2f} %)".format(repr(label) + ",", perc)
        )

    autopct = "%1.1f\\%%" if for_latex else "%1.1f%%"
    pie = ax.pie(percentages, labels=labels, autopct=autopct, **kwargs)
    ax.axis("equal")

    return pie


_kinds = ("fft_as", "pythran", ".pyx", ".py", "built-in", "numpy")


def analyze_stats(
    path, nb_dim=2, plot=False, threshold_long_function=0.01, verbose=True
):
    """Print analysis of profiling result of a 2D solver.

    Parameters
    ----------
    s : pstats.Stats
        Object pointing to a stats file

    """
    stats = pstats.Stats(str(path))
    if verbose:
        stats.sort_stats("time").print_stats(20)

    if nb_dim not in (2, 3):
        raise NotImplementedError

    total_time = 0.0
    for key, value in stats.stats.items():
        time = value[2]
        total_time += time

    long_functions = {}

    key_fft = f"fft{nb_dim}d"
    kinds = _kinds + (key_fft,)
    times = {k: 0.0 for k in kinds}

    for key, value in stats.stats.items():
        name = key[2]
        time = value[2]

        for k in kinds:
            if k in name or k in key[0]:
                if k == ".pyx":
                    if "fft/Sources" in key[0]:
                        continue

                    if "fft_as_arg" in name:
                        continue

                if k == key_fft:
                    if (
                        "__pythran__" in name
                        or "operators.py" in key[0]
                        or "fft_as_arg" in name
                    ):
                        continue

                    callers = value[4]

                    time = 0
                    for kcaller, vcaller in callers.items():
                        if (
                            "fft_as_arg" not in kcaller[2]
                            and "fft_as_arg" not in kcaller[0]
                        ):
                            time += vcaller[2]

                if k == "fft_as" and ".pyx" in key[0]:
                    continue

                if k == ".py" and "fft_as_arg" in name:
                    continue

                if k == "built-in" and "pythran" in name:
                    continue

                times[k] += time

                if time / total_time > threshold_long_function:
                    long_functions[name] = dict(
                        time=time, percentage=100 * time / total_time, kind=k
                    )

    if plot:
        plot_pie(times, long_functions)
        plt.show()

    if not verbose:
        return times, long_functions

    print(
        "\nlong_functions (more than {} % of total time):".format(
            100 * threshold_long_function
        )
    )
    for name, d in long_functions.items():
        print(name + ":")
        for k, v in d.items():
            if k == "percentage":
                print("    " + k + f" = {v:.2f} %")
            else:
                print("    " + k + " = " + repr(v))

    print("\nAnalysis (percentage of total time):")

    keys = list(times.keys())
    keys.sort(key=lambda key: times[key], reverse=True)

    for k in keys:
        t = times[k]
        if t > 0:
            print(
                "time {:10s}: {:7.03f} % ({:4.02f} s)".format(
                    k, t / total_time * 100, t
                )
            )

    print(
        "-" * 26
        + "\n{:15s}  {:7.03f} %".format(
            "", sum([t for t in times.values()]) / total_time * 100
        )
    )

    del times[".py"]
    time_in_not_py = sum([t for t in times.values()])
    print(
        "In not Python functions:\n{:15s}".format("")
        + "  {:7.03f} %".format(time_in_not_py / total_time * 100)
    )

    return times, long_functions
